"""
Implements a selection of the folders and folder items found in an Exchange account.

Exchange is very picky about things like the order of XML elements in SOAP requests, so we need to generate XML
automatically instead of taking advantage of Python SOAP libraries and the WSDL file.
"""

from copy import deepcopy
from logging import getLogger
from xml.etree.cElementTree import Element
import re

from .credentials import DELEGATE
from .ewsdatetime import EWSDateTime, EWSTimeZone
from .restriction import Restriction
from .services import TNS, FindItem, IdOnly, SHALLOW, DEEP, DeleteItem, CreateItem, UpdateItem, FindFolder, GetFolder, \
    GetItem
from .util import set_xml_attr, get_xml_attrs, get_xml_attr, set_xml_value

log = getLogger(__name__)

ElementType = type(Element('x'))  # Type is auto-generated inside cElementTree


class Item:
    ITEM_TYPE = 'Item'
    FIELDURI_PREFIX = 'item'

    # 'extern_id' is not a native EWS Item field. We use it for identification when item originates in an external
    # system. The field is implemented as an extended property on the Item.
    ITEM_FIELDS = (
        'item_id',
        'changekey',
        'subject',
        'body',
        'reminder_is_set',
        'categories',
        'extern_id',
    )
    # These are optional
    EXTRA_ITEM_FIELDS = (
        'datetime_created',
        'datetime_sent',
        'datetime_recieved',
        'last_modified_name',
        'last_modified_time',
    )
    ATTR_FIELDURI_MAP = {
        'subject': 'Subject',
        'body': 'Body',
        'reminder_is_set': 'ReminderIsSet',
        'categories': 'Categories',
        'datetime_created': 'DateTimeCreated',
        'datetime_sent': 'DateTimeSent',
        'datetime_recieved': 'DateTimeReceived',
        'last_modified_name': 'LastModifiedName',
        'last_modified_time': 'LastModifiedTime',
        'extern_id': dict(PropertySetId='c11ff724-aa03-4555-9952-8fa248a11c3e', PropertyName='External ID',
                          PropertyType='String'),
    }
    FIELD_TYPE_MAP = {
        'subject': str,
        'body': str,
        'reminder_is_set': bool,
        'categories': list,
        'datetime_created': EWSDateTime,
        'datetime_sent': EWSDateTime,
        'datetime_recieved': EWSDateTime,
        'last_modified_name': str,
        'last_modified_time': EWSDateTime,
        'extern_id': dict,
    }

    def __init__(self, **kwargs):
        for k in Item.ITEM_FIELDS + Item.EXTRA_ITEM_FIELDS:
            default = False if k == 'reminder_is_set' else None
            setattr(self, k, kwargs.pop(k, default))
        for k, v in kwargs.items():
            raise TypeError("'%s' is an invalid keyword argument for this function" % k)

    @classmethod
    def fieldnames(cls, with_extra=False):
        # Return non-ID field names
        if with_extra:
            return cls.ITEM_FIELDS[2:] + cls.EXTRA_ITEM_FIELDS
        return cls.ITEM_FIELDS[2:]

    @classmethod
    def fielduri_for_field(cls, fieldname):
        try:
            field_uri = cls.ATTR_FIELDURI_MAP[fieldname]
            if isinstance(field_uri, str):
                return '%s:%s' % (cls.FIELDURI_PREFIX, field_uri)
            return field_uri
        except KeyError:
            raise ValueError("No fielduri defined for fieldname '%s'" % fieldname)

    @classmethod
    def request_xml_elem_for_field(cls, fieldname):
        try:
            field_uri = cls.ATTR_FIELDURI_MAP[fieldname]
            if isinstance(field_uri, dict):
                extended_property = Element('t:ExtendedProperty')
                extended_property_field_uri = Element('t:ExtendedFieldURI', **field_uri)
                extended_property.append(extended_property_field_uri)
                return extended_property
            else:
                elem = Element('t:%s' % field_uri)
                if fieldname == 'body':
                    elem.set('BodyType', 'Text')
                return elem
        except KeyError:
            raise ValueError("No fielduri defined for fieldname '%s'" % fieldname)

    @classmethod
    def response_xml_elem_for_field(cls, fieldname):
        try:
            return '{%s}%s' % (TNS, cls.ATTR_FIELDURI_MAP[fieldname])
        except KeyError:
            raise ValueError("No fielduri defined for fieldname '%s'" % fieldname)

    @classmethod
    def type_for_field(cls, fieldname):
        try:
            return cls.FIELD_TYPE_MAP[fieldname]
        except KeyError:
            raise ValueError("No type defined for fieldname '%s'" % fieldname)

    @classmethod
    def additional_property_fields(cls, with_extra=False):
        fields = []
        fielduri = Element('t:FieldURI')
        extended_fielduri = Element('t:ExtendedFieldURI')
        for f in cls.fieldnames(with_extra=with_extra):
            field_uri = cls.fielduri_for_field(f)
            if isinstance(field_uri, str):
                f = deepcopy(fielduri)
                f.set('FieldURI', field_uri)
                fields.append(f)
            else:
                f = deepcopy(extended_fielduri)
                for k, v in field_uri.items():
                    f.set(k, v)
                fields.append(f)
        return fields

    @classmethod
    def id_from_xml(cls, elem):
        id_elem = elem.find('{%s}ItemId' % TNS)
        return id_elem.get('Id'), id_elem.get('ChangeKey')

    @classmethod
    def from_xml(cls, elem, with_extra=False):
        item_id, changekey = cls.id_from_xml(elem)
        kwargs = {}
        extended_properties = elem.findall('{%s}ExtendedProperty' % TNS)
        for fieldname in cls.fieldnames(with_extra=with_extra):
            t = cls.type_for_field(fieldname)
            if t == EWSDateTime:
                str_val = get_xml_attr(elem, cls.response_xml_elem_for_field(fieldname))
                kwargs[fieldname] = EWSDateTime.from_string(str_val)
            elif t == bool:
                val = get_xml_attr(elem, cls.response_xml_elem_for_field(fieldname))
                kwargs[fieldname] = True if val == 'true' else False
            elif t == str:
                kwargs[fieldname] = get_xml_attr(elem, cls.response_xml_elem_for_field(fieldname)) or ''
            elif t == list:
                iter_elem = elem.find(cls.response_xml_elem_for_field(fieldname))
                kwargs[fieldname] = get_xml_attrs(iter_elem, '{%s}String' % TNS) if iter_elem is not None else None
            elif t == dict:
                field_uri = cls.fielduri_for_field(fieldname)
                extended_field_value = None
                for e in extended_properties:
                    extended_field_uri = e.find('{%s}ExtendedFieldURI' % TNS)
                    match = True
                    for k, v in field_uri.items():
                        if extended_field_uri.get(k) != v:
                            match = False
                            break
                    if match:
                        extended_field_value = get_xml_attr(e, '{%s}Value' % TNS) or ''
                        break
                kwargs[fieldname] = extended_field_value
            else:
                assert False, 'Field %s type %s not supported' % (fieldname, t)
        return cls(item_id=item_id, changekey=changekey, **kwargs)


class Folder:
    DISTINGUISHED_FOLDER_ID = None
    CONTAINER_CLASS = None  # See http://msdn.microsoft.com/en-us/library/hh354773(v=exchg.80).aspx
    item_model = Item

    def __init__(self, account, name=None, folder_class=None, folder_id=None, changekey=None):
        self.account = account
        self.name = name or self.DISTINGUISHED_FOLDER_ID
        self.folder_class = folder_class
        self.folder_id = folder_id
        self.changekey = changekey
        if not self.is_distinguished:
            assert self.folder_id
        if self.folder_id:
            assert self.changekey
        self.with_extra_fields = False
        log.debug('%s created for %s', self.__class__.__name__, account)

    @property
    def is_distinguished(self):
        return self.name.lower() == self.DISTINGUISHED_FOLDER_ID

    @classmethod
    def attr_to_fielduri(cls, fieldname):
        return cls.item_model.fielduri_for_field(fieldname)

    @classmethod
    def attr_to_request_xml_elem(cls, fieldname):
        return cls.item_model.request_xml_elem_for_field(fieldname)

    @classmethod
    def attr_to_response_xml_elem(cls, fieldname):
        return cls.item_model.response_xml_elem_for_field(fieldname)

    def find_items(self, start=None, end=None, categories=None, shape=IdOnly, depth=SHALLOW):
        """
        Finds all items in the folder, optionally restricted by start- and enddates and a list of categories
        """
        log.debug(
            'Finding %s items for %s from %s to %s with cats %s shape %s',
            self.DISTINGUISHED_FOLDER_ID,
            self.account,
            start,
            end,
            categories,
            shape
        )
        xml_func = self.item_model.id_from_xml if shape == IdOnly else self.item_model.from_xml
        # Define the extra properties we want on the return objects. 'body' field can only be fetched with GetItem.
        additional_fields = ['item:Categories'] if categories else None
        # Define any search restrictions we want to set on the search.
        # TODO Filtering by category doesn't work on Exchange 2010, returning "ErrorContainsFilterWrongType:
        # The Contains filter can only be used for string properties." Fall back to filtering after getting all items
        # instead. This may be a legal problem because we get ALL items, including private appointments.
        restriction = Restriction.from_params(self.DISTINGUISHED_FOLDER_ID, start=start, end=end)
        items = FindItem(self.account.protocol).call(folder=self, additional_fields=additional_fields,
                                                     restriction=restriction, shape=shape, depth=depth)
        if not categories:
            log.debug('Found %s items', len(items))
            return list(map(xml_func, items))

        # Filter for category. Searching for categories only works with 'Or' operator on Exchange 2007, so we need to
        # ignore items with only some but not all categories present.
        filtered_items = []
        categoryset = set(categories)
        for item in items:
            if not isinstance(item, ElementType):
                status, error = item
                assert not status
                log.warning('Error fetching items: %s', error)
                continue
            cats = item.find(self.attr_to_response_xml_elem('categories'))
            if cats is not None:
                item_cats = get_xml_attrs(cats, '{%s}String' % TNS)
                if categoryset.issubset(set(item_cats)):
                    filtered_items.append(item)
        log.debug('%s of %s items match category', len(filtered_items), len(items))
        return list(map(xml_func, filtered_items))

    def add_items(self, items):
        """
        Creates new items in the folder. 'items' is an iterable of Item objects. Returns a list of (id, changekey)
        tuples in the same order as the input.
        """
        assert len(items)
        log.debug('Adding %s calendar items', len(items))
        return list(map(self.item_model.id_from_xml, CreateItem(self.account.protocol).call(folder=self, items=items)))

    def delete_items(self, ids):
        """
        Deletes items in the folder. 'ids' is an iterable of either (item_id, changekey) tuples or Item objects.
        """
        assert len(ids)
        return DeleteItem(self.account.protocol).call(folder=self, ids=ids)

    def update_items(self, items):
        """
        Updates items in the folder. 'items' is an iterable of tuples containing two elements:

            1. either an (item_id, changekey) tuple or an Item object
            2. a dict containing the Item attributes to change

        """
        assert len(items)
        return list(map(self.item_model.id_from_xml, UpdateItem(self.account.protocol).call(folder=self, items=items)))

    def get_items(self, ids):
        # get_xml() uses self.with_extra_fields. Pass this to from_xml()
        assert len(ids)
        return list(map(
            lambda i: self.item_model.from_xml(i, self.with_extra_fields),
            GetItem(self.account.protocol).call(folder=self, ids=ids))
        )

    def test_access(self):
        """
        Does a simple FindItem to test (read) access to the mailbox. Maybe the account doesn't exist, maybe the
        service user doesn't have access to the calendar.
        """
        now = EWSDateTime.now(tz=EWSTimeZone.timezone('UTC'))
        restriction = Restriction.from_params(self.DISTINGUISHED_FOLDER_ID, start=now, end=now)
        FindItem(self.account.protocol).call(folder=self, restriction=restriction, shape=IdOnly)
        return True

    def folderid_xml(self):
        if self.folder_id:
            assert self.changekey
            return Element('t:FolderId', Id=self.folder_id, ChangeKey=self.changekey)
        else:
            # Only use distinguished ID if we don't have the folder ID
            distinguishedfolderid = Element('t:DistinguishedFolderId', Id=self.DISTINGUISHED_FOLDER_ID)
            if self.account.access_type == DELEGATE:
                mailbox = Element('t:Mailbox')
                set_xml_attr(mailbox, 't:EmailAddress', self.account.primary_smtp_address)
                distinguishedfolderid.append(mailbox)
            return distinguishedfolderid

    def get_xml(self, ids):
        assert len(ids)
        assert isinstance(ids[0], (tuple, Item))
        # This list should be configurable. 'body' element can only be fetched with GetItem.
        # CalendarItem.from_xml() specifies the items we currently expect. For full list, see
        # https://msdn.microsoft.com/en-us/library/office/aa494315(v=exchg.150).aspx
        log.debug(
            'Getting %s %s items for %s',
            len(ids),
            self.DISTINGUISHED_FOLDER_ID,
            self.account
        )
        getitem = Element('m:%s' % GetItem.SERVICE_NAME)
        itemshape = Element('m:ItemShape')
        set_xml_attr(itemshape, 't:BaseShape', IdOnly)
        additional_properties = self.item_model.additional_property_fields(with_extra=self.with_extra_fields)
        if additional_properties:
            additionalproperties = Element('t:AdditionalProperties')
            for p in additional_properties:
                additionalproperties.append(p)
            itemshape.append(additionalproperties)
        getitem.append(itemshape)
        itemids = Element('m:ItemIds')
        itemid = Element('t:ItemId')
        # copy.deepcopy() is an order of magnitude faster than having
        # Element() inside the loop. It matters because 'ids' may
        # be a very large list.
        for item in ids:
            item_id, changekey = item if isinstance(item, tuple) else (item.item_id, item.changekey)
            i = deepcopy(itemid)
            i.set('Id', item_id)
            i.set('ChangeKey', changekey)
            itemids.append(i)
        getitem.append(itemids)
        return getitem

    def create_xml(self, items):
        assert len(items)
        # Takes an account name, a folder name, a list of Calendar.Item obejcts and a function to convert items to XML
        # Elements
        if isinstance(self, Calendar):
            createitem = Element('m:%s' % CreateItem.SERVICE_NAME, SendMeetingInvitations='SendToNone')
        elif isinstance(self, Inbox):
            createitem = Element('m:%s' % CreateItem.SERVICE_NAME, MessageDisposition='SaveOnly')
        else:
            createitem = Element('m:%s' % CreateItem.SERVICE_NAME)

        saveditemfolderid = Element('m:SavedItemFolderId')
        saveditemfolderid.append(self.folderid_xml())
        createitem.append(saveditemfolderid)

        myitems = Element('m:Items')
        for item in self._create_item_xml(items):
            myitems.append(item)
        createitem.append(myitems)
        return createitem

    def _create_item_xml(self, items):
        raise NotImplementedError()

    def delete_xml(self, ids):
        assert len(ids)
        assert isinstance(ids[0], (tuple, Item))
        # Prepare reuseable Element objects
        if isinstance(self, Calendar):
            deleteitem = Element('m:%s' % DeleteItem.SERVICE_NAME, DeleteType='HardDelete',
                                 SendMeetingCancellations='SendToNone')
        elif isinstance(self, Inbox):
            deleteitem = Element('m:%s' % DeleteItem.SERVICE_NAME, DeleteType='HardDelete')
        elif isinstance(self, Tasks):
            deleteitem = Element('m:%s' % DeleteItem.SERVICE_NAME, DeleteType='HardDelete',
                                 AffectedTaskOccurrences='SpecifiedOccurrenceOnly')
        else:
            deleteitem = Element('m:%s' % DeleteItem.SERVICE_NAME, DeleteType='HardDelete')
        if self.account.version.major_version >= 15:
            deleteitem.set('SuppressReadReceipts', 'true')

        itemids = Element('m:ItemIds')
        itemid = Element('t:ItemId')

        # copy.deepcopy() is an order of magnitude faster than having
        # Element() inside the loop. It matters because 'ids' may
        # be a very large list.
        for item in ids:
            item_id, changekey = item if isinstance(item, tuple) else (item.item_id, item.changekey)
            i = deepcopy(itemid)
            i.set('Id', item_id)
            i.set('ChangeKey', changekey)
            itemids.append(i)
        deleteitem.append(itemids)
        return deleteitem

    def update_xml(self, items):
        assert len(items)
        assert isinstance(items[0], tuple)
        assert isinstance(items[0][0], (tuple, Item))
        assert isinstance(items[0][1], dict)
        # Prepare reuseable Element objects
        if isinstance(self, Calendar):
            updateitem = Element('m:%s' % UpdateItem.SERVICE_NAME, ConflictResolution='AutoResolve',
                                 SendMeetingInvitationsOrCancellations='SendToNone')
        elif isinstance(self, Inbox):
            updateitem = Element('m:%s' % UpdateItem.SERVICE_NAME, ConflictResolution='AutoResolve',
                                 MessageDisposition='SaveOnly')
        else:
            updateitem = Element('m:%s' % UpdateItem.SERVICE_NAME, ConflictResolution='AutoResolve')
        if self.account.version.major_version >= 15:
            updateitem.set('SuppressReadReceipts', 'true')

        itemchanges = Element('m:ItemChanges')
        itemchange = Element('t:ItemChange')
        itemid = Element('t:ItemId')
        updates = Element('t:Updates')
        setitemfield = Element('t:SetItemField')
        deleteitemfield = Element('t:DeleteItemField')
        fielduri = Element('t:FieldURI')
        extended_fielduri = Element('t:ExtendedFieldURI')
        folderitem = Element('t:%s' % self.item_model.ITEM_TYPE)

        # copy.deepcopy() is an order of magnitude faster than having
        # Element() inside the loop. It matters because 'ids' may
        # be a very large list.
        for item, update_dict in items:
            assert len(update_dict)
            i_itemchange = deepcopy(itemchange)
            item_id, changekey = item if isinstance(item, tuple) else (item.item_id, item.changekey)
            i_itemid = deepcopy(itemid)
            i_itemid.set('Id', item_id)
            i_itemid.set('ChangeKey', changekey)
            i_itemchange.append(i_itemid)
            i_updates = deepcopy(updates)
            meeting_timezone_added = False
            for fieldname, val in update_dict.items():
                field_uri = self.attr_to_fielduri(fieldname)
                if isinstance(field_uri, str):
                    i_fielduri = deepcopy(fielduri)
                    i_fielduri.set('FieldURI', field_uri)
                else:
                    i_fielduri = deepcopy(extended_fielduri)
                    for k, v in field_uri.items():
                        i_fielduri.set(k, v)
                if val is None:
                    # A value of None means we want to remove this field from the item
                    i_deleteitemfield = deepcopy(deleteitemfield)
                    i_deleteitemfield.append(i_fielduri)
                    i_updates.append(i_deleteitemfield)
                    continue
                i_setitemfield = deepcopy(setitemfield)
                i_setitemfield.append(i_fielduri)
                i_folderitem = deepcopy(folderitem)
                i_value = self.attr_to_request_xml_elem(fieldname)
                if isinstance(val, EWSDateTime):
                    i_folderitem.append(set_xml_value(i_value, val))
                    i_setitemfield.append(i_folderitem)
                    i_updates.append(i_setitemfield)

                    # Always set timezone explicitly when updating date fields. Exchange 2007 wants "MeetingTimeZone"
                    # instead of explicit timezone on each datetime field.
                    if self.account.version.major_version < 14:
                        if meeting_timezone_added:
                            # Let's hope that we're not changing timezone, or that both 'start' and 'end' are supplied.
                            # Exchange 2007 doesn't support different timezone on start and end.
                            continue
                        i_setitemfield_tz = deepcopy(setitemfield)
                        i_fielduri = deepcopy(fielduri)
                        i_fielduri.set('FieldURI', 'calendar:MeetingTimeZone')
                        i_setitemfield_tz.append(i_fielduri)
                        i_folderitem = deepcopy(folderitem)
                        meeting_timezone = Element('t:MeetingTimeZone', TimeZoneName=val.tzinfo.ms_id)
                        i_folderitem.append(meeting_timezone)
                        i_setitemfield_tz.append(i_folderitem)
                        i_updates.append(i_setitemfield_tz)
                        meeting_timezone_added = True
                    else:
                        i_setitemfield_tz = deepcopy(setitemfield)
                        i_fielduri = deepcopy(fielduri)
                        if fieldname == 'start':
                            field_uri = 'calendar:StartTimeZone'
                            timezone_element = Element('t:StartTimeZone', Id=val.tzinfo.ms_id, Name=val.tzinfo.ms_name)
                        elif fieldname == 'end':
                            field_uri = 'calendar:EndTimeZone'
                            timezone_element = Element('t:EndTimeZone', Id=val.tzinfo.ms_id, Name=val.tzinfo.ms_name)
                        else:
                            assert False, 'Cannot set timezone for field %s' % fieldname
                        i_fielduri.set('FieldURI', field_uri)
                        i_setitemfield_tz.append(i_fielduri)
                        i_folderitem = deepcopy(folderitem)
                        i_folderitem.append(timezone_element)
                        i_setitemfield_tz.append(i_folderitem)
                        i_updates.append(i_setitemfield_tz)
                else:
                    if isinstance(field_uri, str):
                        i_folderitem.append(set_xml_value(i_value, val))
                    else:
                        i_folderitem.append(set_xml_attr(i_value, 't:Value', val))
                    i_setitemfield.append(i_folderitem)
                    i_updates.append(i_setitemfield)
            i_itemchange.append(i_updates)
            itemchanges.append(i_itemchange)
        updateitem.append(itemchanges)
        return updateitem

    @classmethod
    def from_xml(cls, account, elem):
        fld_type = re.sub('{.*}', '', elem.tag)
        fld_class = FOLDER_TYPE_MAP[fld_type]
        fld_id_elem = elem.find('{%s}FolderId' % TNS)
        fld_id = fld_id_elem.get('Id')
        changekey = fld_id_elem.get('ChangeKey')
        display_name = get_xml_attr(elem, '{%s}DisplayName' % TNS)
        folder_class = get_xml_attr(elem, '{%s}FolderClass' % TNS)
        return fld_class(account=account, name=display_name, folder_class=folder_class, folder_id=fld_id,
                         changekey=changekey)

    def get_folders(self, shape=IdOnly, depth=DEEP):
        folders = []
        for elem in FindFolder(self.account.protocol).call(
                folder=self,
                additional_fields=['folder:DisplayName', 'folder:FolderClass'],
                shape=shape,
                depth=depth
        ):
            folders.append(self.from_xml(self.account, elem))
        return folders

    def get_folder(self, shape=IdOnly):
        folders = []
        for elem in GetFolder(self.account.protocol).call(
                folder=self,
                additional_fields=['folder:DisplayName', 'folder:FolderClass'],
                shape=shape
        ):
            folders.append(self.from_xml(self.account, elem))
        assert len(folders) == 1
        return folders[0]

    def __repr__(self):
        return self.__class__.__name__ + \
               repr((self.account, self.name, self.folder_class, self.folder_id, self.changekey))

    def __str__(self):
        return '%s (%s)' % (self.__class__.__name__, self.name)


class Root(Folder):
    DISTINGUISHED_FOLDER_ID = 'root'


class CalendarItem(Item):
    """
    Models a calendar item. Not all attributes are supported. See full list at
    https://msdn.microsoft.com/en-us/library/office/aa564765(v=exchg.150).aspx
    """
    ITEM_TYPE = 'CalendarItem'
    SUBJECT_MAXLENGTH = 255
    LOCATION_MAXLENGTH = 255
    FIELDURI_PREFIX = 'calendar'
    ITEM_FIELDS = (
        'start',
        'end',
        'location',
        'organizer',
        'legacy_free_busy_status',
    )
    ATTR_FIELDURI_MAP = {
        'start': 'Start',
        'end': 'End',
        'location': 'Location',
        'organizer': 'Organizer',
        'legacy_free_busy_status': 'LegacyFreeBusyStatus',
    }
    FIELD_TYPE_MAP = {
        'start': EWSDateTime,
        'end': EWSDateTime,
        'location': str,
        'organizer': str,
        'legacy_free_busy_status': str,
    }

    @classmethod
    def fieldnames(cls, with_extra=False):
        return cls.ITEM_FIELDS + Item.fieldnames(with_extra=with_extra)

    @classmethod
    def fielduri_for_field(cls, fieldname):
        try:
            field_uri = cls.ATTR_FIELDURI_MAP[fieldname]
            if isinstance(field_uri, str):
                return '%s:%s' % (cls.FIELDURI_PREFIX, field_uri)
            return field_uri
        except KeyError:
            return Item.fielduri_for_field(fieldname)

    @classmethod
    def request_xml_elem_for_field(cls, fieldname):
        try:
            field_uri = cls.ATTR_FIELDURI_MAP[fieldname]
            if isinstance(field_uri, dict):
                extended_property = Element('t:ExtendedProperty')
                extended_property_field_uri = Element('t:ExtendedFieldURI', **field_uri)
                extended_property.append(extended_property_field_uri)
                return extended_property
            else:
                return Element('t:%s' % field_uri)
        except KeyError:
            return Item.request_xml_elem_for_field(fieldname)

    @classmethod
    def response_xml_elem_for_field(cls, fieldname):
        try:
            return '{%s}%s' % (TNS, cls.ATTR_FIELDURI_MAP[fieldname])
        except KeyError:
            return Item.response_xml_elem_for_field(fieldname)

    @classmethod
    def type_for_field(cls, fieldname):
        try:
            return cls.FIELD_TYPE_MAP[fieldname]
        except KeyError:
            return Item.type_for_field(fieldname)

    def __init__(self, **kwargs):
        for k in self.ITEM_FIELDS:
            default = 'Busy' if k == 'legacy_free_busy_status' else None
            v = kwargs.pop(k, default)
            if k in ('start', 'end') and v and not getattr(v, 'tzinfo'):
                raise ValueError("'%s' must be timezone aware")
            setattr(self, k, v)
        super().__init__(**kwargs)

    def __repr__(self):
        return self.__class__.__name__ + repr(getattr(self, k) for k in (Item.ITEM_FIELDS + self.ITEM_FIELDS))

    def __str__(self):
        return '''\
ItemId: %(item_id)s
Changekey: %(changekey)s
Subject: %(subject)s
Start: %(start)s
End: %(end)s
Location: %(location)s
Body: %(body)s
Has reminder: %(reminder_is_set)s
Categories: %(categories)s
Extern ID: %(extern_id)s''' % self.__dict__


class Calendar(Folder):
    """
    An interface for the Exchange calendar
    """
    DISTINGUISHED_FOLDER_ID = 'calendar'
    CONTAINER_CLASS = 'IPF.Appointment'
    item_model = CalendarItem

    # These must be capitalized
    # TODO: This is most definitely not either complete or authoritative
    LOCALIZED_NAMES = (
        'Kalender',
    )

    def _create_item_xml(self, items):
        """
        Takes an array of Calendar.Item objects and generates the XML.
        """
        calendar_items = []
        # Prepare Element objects for CalendarItem creation
        calendar_item = Element('t:%s' % self.item_model.ITEM_TYPE)
        subject = self.attr_to_request_xml_elem('subject')
        body = self.attr_to_request_xml_elem('body')
        reminder = self.attr_to_request_xml_elem('reminder_is_set')
        start = self.attr_to_request_xml_elem('start')
        end = self.attr_to_request_xml_elem('end')
        busystatus = self.attr_to_request_xml_elem('legacy_free_busy_status')
        location = self.attr_to_request_xml_elem('location')
        organizer = self.attr_to_request_xml_elem('organizer')
        categories = self.attr_to_request_xml_elem('categories')
        extern_id = self.attr_to_request_xml_elem('extern_id')

        # Using copy.deepcopy() increases performance drastically on large arrays
        for item in items:
            # WARNING: The order of addition of XML elements is VERY important. Exchange expects XML elements in a
            # specific, non-documented order and will fail with meaningless errors if the order is wrong.
            i = deepcopy(calendar_item)
            i.append(set_xml_value(deepcopy(subject), item.subject))
            if item.body:
                i.append(set_xml_value(deepcopy(body), item.body))
            if item.categories:
                i.append(set_xml_value(deepcopy(categories), item.categories))
            i.append(set_xml_value(deepcopy(reminder), item.reminder_is_set))
            if item.extern_id is not None:
                i.append(set_xml_attr(deepcopy(extern_id), 't:Value', item.extern_id))
            i.append(set_xml_value(deepcopy(start), item.start))
            i.append(set_xml_value(deepcopy(end), item.end))
            i.append(set_xml_value(deepcopy(busystatus), item.legacy_free_busy_status))
            if item.location:
                i.append(set_xml_value(deepcopy(location), item.location))
            if item.organizer:
                i.append(set_xml_value(deepcopy(organizer), item.organizer))
            if self.account.version.major_version < 14:
                meeting_timezone = Element('t:MeetingTimeZone', TimeZoneName=item.start.tzinfo.ms_id)
                i.append(meeting_timezone)
            else:
                start_timezone = Element('t:StartTimeZone', Id=item.start.tzinfo.ms_id, Name=item.start.tzinfo.ms_name)
                i.append(start_timezone)
                end_timezone = Element('t:EndTimeZone', Id=item.end.tzinfo.ms_id, Name=item.end.tzinfo.ms_name)
                i.append(end_timezone)
            calendar_items.append(i)
        return calendar_items


class Message(Item):
    ITEM_TYPE = 'Message'


class Inbox(Folder):
    ITEM_TYPE = 'Message'
    DISTINGUISHED_FOLDER_ID = 'inbox'
    CONTAINER_CLASS = 'IPF.Note'


class Task(Item):
    ITEM_TYPE = 'Task'


class Tasks(Folder):
    DISTINGUISHED_FOLDER_ID = 'tasks'
    CONTAINER_CLASS = 'IPF.Task'


class Contact(Item):
    ITEM_TYPE = 'Contact'


class Contacts(Folder):
    DISTINGUISHED_FOLDER_ID = 'contacts'
    CONTAINER_CLASS = 'IPF.Contact'


class GenericFolder(Folder):
    pass


class WellknownFolder(Folder):
    # Use this class until we have specific folder implementations
    pass


# See http://msdn.microsoft.com/en-us/library/microsoft.exchange.webservices.data.wellknownfoldername(v=exchg.80).aspx
WELLKNOWN_FOLDERS = dict([
    ('Calendar', Calendar),
    ('Contacts', Contacts),
    ('DeletedItems', WellknownFolder),
    ('Drafts', WellknownFolder),
    ('Inbox', Inbox),
    ('Journal', WellknownFolder),
    ('Notes', WellknownFolder),
    ('Outbox', WellknownFolder),
    ('SentItems', WellknownFolder),
    ('Tasks', Tasks),
    ('MsgFolderRoot', WellknownFolder),
    ('PublicFoldersRoot', WellknownFolder),
    ('Root', Root),
    ('JunkEmail', WellknownFolder),
    ('Search', WellknownFolder),
    ('VoiceMail', WellknownFolder),
    ('RecoverableItemsRoot', WellknownFolder),
    ('RecoverableItemsDeletions', WellknownFolder),
    ('RecoverableItemsVersions', WellknownFolder),
    ('RecoverableItemsPurges', WellknownFolder),
    ('ArchiveRoot', WellknownFolder),
    ('ArchiveMsgFolderRoot', WellknownFolder),
    ('ArchiveDeletedItems', WellknownFolder),
    ('ArchiveRecoverableItemsRoot', Folder),
    ('ArchiveRecoverableItemsDeletions', WellknownFolder),
    ('ArchiveRecoverableItemsVersions', WellknownFolder),
    ('ArchiveRecoverableItemsPurges', WellknownFolder),
    ('SyncIssues', WellknownFolder),
    ('Conflicts', WellknownFolder),
    ('LocalFailures', WellknownFolder),
    ('ServerFailures', WellknownFolder),
    ('RecipientCache', WellknownFolder),
    ('QuickContacts', WellknownFolder),
    ('ConversationHistory', WellknownFolder),
    ('ToDoSearch', WellknownFolder),
    ('', GenericFolder),
])

FOLDER_TYPE_MAP = dict()
for folder_name, folder_model in WELLKNOWN_FOLDERS.items():
    folder_type = '%sFolder' % folder_name
    FOLDER_TYPE_MAP[folder_type] = folder_model



import re
from xml.etree.cElementTree import Element, fromstring, ParseError
import logging
from io import StringIO
import time
from threading import get_ident
from datetime import datetime
from socket import timeout as SocketTimeout
from urllib.parse import urlparse

from lxml.etree import XMLParser, parse, tostring
from requests.exceptions import ConnectionError, ReadTimeout, ChunkedEncodingError

from .errors import TransportError, RateLimitError, RedirectError


log = logging.getLogger(__name__)

# Some control characters are illegal in XML 1.0 (and XML 1.1). Some Exchange serveres will emit XML 1.0
# containing characters only allowed in XML 1.1
_illegal_xml_chars_RE = re.compile('[\x00-\x08\x0b\x0c\x0e-\x1F\uD800-\uDFFF\uFFFE\uFFFF]')
BOM = '\xef\xbb\xbf'  # UTF-8 byte order mark which tails some Microsoft XML strings


def chunkify(iterable, chunksize):
    """
    Splits an iterable into chunks of size ``chunksize``. The last chunk may be smaller than ``chunksize``.
    """
    for i in range(0, len(iterable), chunksize):
        yield iterable[i:i + chunksize]


def get_xml_attr(tree, name):
    elem = tree.find(name)
    if elem is not None and elem.text:  # Must compare with None, see XML docs
        return elem.text.strip() or None
    return None


def get_xml_attrs(tree, name):
    return [elem.text.strip() for elem in tree.findall(name) if elem.text is not None]


def set_xml_value(elem, value):
    from .ewsdatetime import EWSDateTime
    if isinstance(value, str):
        elem.text = safe_xml_value(value)
    elif isinstance(value, (tuple, list)):
        for v in value:
            set_xml_attr(elem, 't:String', v)
    elif isinstance(value, bool):
        elem.text = '1' if value else '0'
    elif isinstance(value, EWSDateTime):
        elem.text = value.ewsformat()
    else:
        raise AttributeError('Unsupported type %s for value %s' % (type(value), value))
    return elem


def safe_xml_value(value, replacement='?'):
    return str(_illegal_xml_chars_RE.sub(replacement, value))


def set_xml_attr(tree, name, value):
    elem = Element(name)
    elem.text = safe_xml_value(value)
    tree.append(elem)
    return tree


def to_xml(text, encoding):
    processed = text.lstrip(BOM).encode(encoding or 'utf-8')
    try:
        return fromstring(processed)
    except ParseError:
        # Exchange servers may spit out the weirdest XML. lxml is pretty good at recovering from errors
        log.warning('Fallback to lxml processing of faulty XML')
        magical_parser = XMLParser(encoding=encoding or 'utf-8', recover=True)
        root = parse(StringIO(processed), magical_parser)
        try:
            return fromstring(tostring(root))
        except ParseError as e:
            line_no, col_no = e.lineno, e.offset
            try:
                offending_line = processed.splitlines()[line_no - 1]
            except IndexError:
                offending_line = ''
            offending_excerpt = offending_line[max(0, col_no - 20):col_no + 20].decode('ascii', 'ignore')
            raise ParseError('%s\nOffending text: [...]%s[...]' % (str(e), offending_excerpt)) from e


def is_xml(text):
    """
    Helper function. Lightweight test if response is an XML doc
    """
    return text.lstrip(BOM)[0:5] == '<?xml'


class DummyRequest:
    headers = {}


class DummyResponse:
    status_code = 401
    headers = {}
    text = ''
    request = DummyRequest()


def get_redirect_url(response, server=None, has_ssl=None):
    redirect_url = response.headers.get('location', None)
    if not redirect_url:
        raise TransportError('302 redirect but no location header')
    # At least some are kind enough to supply a new location
    url = urlparse(redirect_url)
    response_url = urlparse(response.url)
    if server is None:
        server = response_url.netloc
    if has_ssl is None:
        has_ssl = response_url.scheme == 'https'
    scheme = url.scheme or ('https' if has_ssl else 'http')
    has_ssl = scheme == 'https'
    if url.netloc:
        server = url.netloc
    server = server.lower()
    redirect_url = '%s://%s%s' % (scheme, server, url.path)
    if redirect_url == response.url:
        # And some are mean enough to redirect to the same location
        raise TransportError('Redirect to same location: %s' % redirect_url)
    return redirect_url, server, has_ssl


def post_ratelimited(protocol, session, url, headers, data, timeout=None, verify=True, allow_redirects=False):
    # The contract on sessions here is to return the session that ends up being used, or retiring the session if we
    # intend to raise an exception. We give up on max_wait timeout, not number of retries
    r = None
    wait = 10  # seconds
    max_wait = 3600  # seconds
    redirects = 0
    max_redirects = 5  # We don't want to be sent into an endless redirect loop
    log_msg = '''\
Retry: %(i)s
Waited: %(wait)s
Timeout: %(timeout)s
Session: %(session_id)s
Thread: %(thread_id)s
Auth type: %(auth)s
URL: %(url)s
Verify: %(verify)s
Allow redirects: %(allow_redirects)s
Response time: %(response_time)s
Status code: %(status_code)s
Request headers: %(request_headers)s
Response headers: %(response_headers)s'''
    log_vals = dict(i=0, wait=0, timeout=timeout, session_id=session.session_id, thread_id=get_ident(),
                    auth=session.auth, url=url, response_time=None, status_code=None, request_headers=headers,
                    response_headers=None, verify=verify, allow_redirects=allow_redirects)
    try:
        while True:
            log.debug('Session %(session_id)s thread %(thread_id)s: retry %(i)s timeout %(timeout)s POST\'ing to '
                      '%(url)s after %(wait)s s wait', log_vals)
            d1 = datetime.now()
            try:
                r = session.post(url=url, headers=headers, data=data, allow_redirects=False, timeout=timeout,
                                 verify=verify)
            except (ChunkedEncodingError, ConnectionError, ConnectionResetError, ReadTimeout, SocketTimeout):
                log.debug(
                    'Session %(session_id)s thread %(thread_id)s: timeout or connection error POST\'ing to %(url)s',
                    log_vals)
                r = DummyResponse()
                r.request.headers = headers
                r.headers = {'DummyResponseHeader': None}
            d2 = datetime.now()
            log_vals['response_time'] = str(d2 - d1)
            log_vals['status_code'] = r.status_code
            log_vals['request_headers'] = r.request.headers
            log_vals['response_headers'] = r.headers
            log.debug(log_msg, log_vals)
            # The genericerrorpage.htm/internalerror.asp is ridiculous behaviour for random outages. Redirect to
            # '/internalsite/internalerror.asp' or '/internalsite/initparams.aspx' is caused by e.g. SSL certificate
            # f*ckups on the Exchange server.
            if (r.status_code == 401) \
                    or (r.headers.get('connection') == 'close') \
                    or (r.status_code == 302 and r.headers.get('location').lower() ==
                        '/ews/genericerrorpage.htm?aspxerrorpath=/ews/exchange.asmx')\
                    or (r.status_code == 503):
                # Maybe stale session. Get brand new one. But wait a bit, since the server may be rate-limiting us.
                # This can be 302 redirect to error page, 401 authentication error or 503 service unavailable
                if r.status_code not in (302, 401, 503):
                    # Only retry if we didn't get a useful response
                    break
                log_vals['i'] += 1
                log_vals['wait'] = wait  # We set it to 0 initially
                if wait > max_wait:
                    # We lost patience. Session is cleaned up in outer loop
                    raise RateLimitError(
                        'Session %(session_id)s URL %(url)s: Max timeout reached' % log_vals)
                log.info("Session %(session_id)s thread %(thread_id)s: Connection error on URL %(url)s "
                         "(code %(status_code)s). Cool down %(wait)s secs", log_vals)
                time.sleep(wait)  # Increase delay for every retry
                wait *= 2
                session = protocol.renew_session(session)
                log_vals['wait'] = wait
                log_vals['session_id'] = session.session_id
                continue
            if r.status_code == 302:
                # If we get a normal 302 redirect, requests will issue a GET to that URL. We still want to POST
                redirect_url, server, has_ssl = get_redirect_url(response=r, server=protocol.server,
                                                                 has_ssl=protocol.has_ssl)
                if not allow_redirects:
                    raise TransportError('Redirect not allowed but we were redirected (%s -> %s)' % (url, redirect_url))
                if has_ssl != protocol.has_ssl or server != protocol.server:
                    log.debug("'allow_redirects' only supports relative redirects (%s -> %s)", url, redirect_url)
                    raise RedirectError(url=redirect_url, server=server, has_ssl=has_ssl)
                url = redirect_url
                log_vals['url'] = url
                log.debug('302 Redirected to %s', url)
                redirects += 1
                if redirects > max_redirects:
                    raise TransportError('Max redirect count exceeded')
                continue
            break
    except (RateLimitError, RedirectError) as e:
        log.warning(e.value)
        protocol.retire_session(session)
        raise
    except Exception as e:
        # Let higher layers handle this. Add data for better debugging.
        log_msg = '%(exc_cls)s: %(exc_msg)s\n' + log_msg
        log_vals['exc_cls'] = e.__class__.__name__
        log_vals['exc_msg'] = str(e)
        log_msg += '\nRequest data: %(data)s'
        log_vals['data'] = data
        log_msg += '\nResponse data: %(text)s'
        try:
            log_vals['text'] = r.text
        except (NameError, AttributeError):
            log_vals['text'] = ''
        log.error(log_msg, log_vals)
        protocol.retire_session(session)
        raise
    if r.status_code != 200:
        if r.text and is_xml(r.text):
            # Some genius at Microsoft thinks it's OK to send 500 error messages with valid SOAP response
            log.debug('Got status code %s but trying to parse content anyway', r.status_code)
        else:
            # This could be anything. Let higher layers handle this
            protocol.retire_session(session)
            log_msg += '\nRequest data: %(data)s'
            log_vals['data'] = data
            try:
                log_msg += '\nResponse data: %(text)s'
                log_vals['text'] = r.text
            except (NameError, AttributeError):
                pass
            raise TransportError('Unknown failure\n' + log_msg % log_vals)
    log.debug('Session %(session_id)s thread %(thread_id)s: Useful response from %(url)s', log_vals)
    return r, session



import datetime

import pytz


class EWSDateTime(datetime.datetime):
    """
    Extends the normal datetime implementation to satisfy EWS
    """

    def __new__(cls, *args, **kwargs):
        """
        Inherits datetime and adds extra formatting required by EWS.
        """
        if 'tzinfo' in kwargs:
            # Creating
            raise ValueError('Do not set tzinfo directly. Use EWSTimeZone.localize() instead')
        self = super().__new__(cls, *args, **kwargs)
        return self

    def ewsformat(self):
        """
        ISO 8601 format to satisfy xs:datetime as interpreted by EWS. Example: 2009-01-15T13:45:56Z
        """
        assert self.tzinfo  # EWS datetimes must always be timezone-aware
        if self.tzinfo.zone == 'UTC':
            return self.strftime('%Y-%m-%dT%H:%M:%SZ')
        return self.strftime('%Y-%m-%dT%H:%M:%S')

    @classmethod
    def from_datetime(cls, d):
        dt = cls(d.year, d.month, d.day, d.hour, d.minute, d.second, d.microsecond)
        if d.tzinfo:
            if isinstance(d.tzinfo, EWSTimeZone):
                return d.tzinfo.localize(dt)
            return EWSTimeZone.from_pytz(d.tzinfo).localize(dt)
        return dt

    def astimezone(self, tz=None):
        t = super().astimezone(tz=tz)
        return self.from_datetime(t)  # We want to return EWSDateTime objects

    def __add__(self, other):
        t = super().__add__(other)
        return self.from_datetime(t)  # We want to return EWSDateTime objects

    def __sub__(self, other):
        t = super().__sub__(other)
        if isinstance(t, datetime.timedelta):
            return t
        return self.from_datetime(t)  # We want to return EWSDateTime objects

    @classmethod
    def from_string(cls, date_string):
        # Assume UTC and return timezone-aware EWSDateTime objects
        local_dt = super().strptime(date_string, '%Y-%m-%dT%H:%M:%SZ')
        return EWSTimeZone.from_pytz(pytz.utc).localize(cls.from_datetime(local_dt))

    @classmethod
    def now(cls, tz=None):
        # We want to return EWSDateTime objects
        t = super().now(tz=tz)
        return cls.from_datetime(t)


class EWSTimeZone:
    """
    Represents a timezone as expected by the EWS TimezoneContext / TimezoneDefinition XML element, and returned by
    services.GetServerTimeZones.
    """
    @classmethod
    def from_pytz(cls, tz):
        # pytz timezones are dynamically generated. Subclass the tz.__class__ and add the extra Microsoft timezone
        # labels we need.
        self_cls = type(cls.__name__, (cls, tz.__class__), dict(tz.__class__.__dict__))
        try:
            self_cls.ms_id = cls.PYTZ_TO_MS_MAP[tz.zone]
        except KeyError as e:
            raise ValueError('Please add an entry for "%s" in PYTZ_TO_MS_TZMAP' % tz.zone) from e
        self_cls.ms_name = cls.MS_TIMEZONE_DEFINITIONS[self_cls.ms_id]
        self = self_cls()
        for k, v in tz.__dict__.items():
            setattr(self, k, v)
        return self

    @classmethod
    def timezone(cls, location):
        # Like pytz.timezone() but returning EWSTimeZone instances
        tz = pytz.timezone(location)
        return cls.from_pytz(tz)

    def normalize(self, dt):
        # super() returns a dt.tzinfo of class pytz.tzinfo.FooBar. We need to return type EWSTimeZone
        res = super().normalize(dt)
        return res.replace(tzinfo=self.from_pytz(res.tzinfo))

    def localize(self, dt):
        # super() returns a dt.tzinfo of class pytz.tzinfo.FooBar. We need to return type EWSTimeZone
        res = super().localize(dt)
        return res.replace(tzinfo=self.from_pytz(res.tzinfo))

    # Manually maintained translation between pytz location / timezone name and MS timezone IDs
    PYTZ_TO_MS_MAP = {
        'Europe/Copenhagen': 'Romance Standard Time',
        'UTC': 'UTC',
        'GMT': 'GMT Standard Time',
    }

    # This is a somewhat authoritative list of the timezones available on an Exchange server. Format is (id, name).
    # For a full list supported by the target server, call services.GetServerTimeZones.
    MS_TIMEZONE_DEFINITIONS = dict([
        ('Dateline Standard Time', '(UTC-12:00) International Date Line West'),
        ('UTC-11', '(UTC-11:00) Coordinated Universal Time-11'),
        ('Samoa Standard Time', '(UTC-11:00) Midway Island, Samoa'),
        ('Hawaiian Standard Time', '(UTC-10:00) Hawaii'),
        ('Alaskan Standard Time', '(UTC-09:00) Alaska'),
        ('Pacific Standard Time', '(UTC-08:00) Pacific Time (US & Canada)'),
        ('Pacific Standard Time (Mexico)', '(UTC-08:00) Tijuana, Baja California'),
        ('US Mountain Standard Time', '(UTC-07:00) Arizona'),
        ('Mountain Standard Time (Mexico)', '(UTC-07:00) Chihuahua, La Paz, Mazatlan'),
        ('Mountain Standard Time', '(UTC-07:00) Mountain Time (US & Canada)'),
        ('Central America Standard Time', '(UTC-06:00) Central America'),
        ('Central Standard Time', '(UTC-06:00) Central Time (US & Canada)'),
        ('Central Standard Time (Mexico)', '(UTC-06:00) Guadalajara, Mexico City, Monterrey'),
        ('Canada Central Standard Time', '(UTC-06:00) Saskatchewan'),
        ('SA Pacific Standard Time', '(UTC-05:00) Bogota, Lima, Quito'),
        ('Eastern Standard Time', '(UTC-05:00) Eastern Time (US & Canada)'),
        ('US Eastern Standard Time', '(UTC-05:00) Indiana (East)'),
        ('Venezuela Standard Time', '(UTC-04:30) Caracas'),
        ('Paraguay Standard Time', '(UTC-04:00) Asuncion'),
        ('Atlantic Standard Time', '(UTC-04:00) Atlantic Time (Canada)'),
        ('SA Western Standard Time', '(UTC-04:00) Georgetown, La Paz, San Juan'),
        ('Central Brazilian Standard Time', '(UTC-04:00) Manaus'),
        ('Pacific SA Standard Time', '(UTC-04:00) Santiago'),
        ('Newfoundland Standard Time', '(UTC-03:30) Newfoundland'),
        ('E. South America Standard Time', '(UTC-03:00) Brasilia'),
        ('Argentina Standard Time', '(UTC-03:00) Buenos Aires'),
        ('SA Eastern Standard Time', '(UTC-03:00) Cayenne'),
        ('Greenland Standard Time', '(UTC-03:00) Greenland'),
        ('Montevideo Standard Time', '(UTC-03:00) Montevideo'),
        ('UTC-02', '(UTC-02:00) Coordinated Universal Time-02'),
        ('Mid-Atlantic Standard Time', '(UTC-02:00) Mid-Atlantic'),
        ('Azores Standard Time', '(UTC-01:00) Azores'),
        ('Cape Verde Standard Time', '(UTC-01:00) Cape Verde Is.'),
        ('Morocco Standard Time', '(UTC) Casablanca'),
        ('UTC', '(UTC) Coordinated Universal Time'),
        ('GMT Standard Time', '(UTC) Greenwich Mean Time : Dublin, Edinburgh, Lisbon, London'),
        ('Greenwich Standard Time', '(UTC) Monrovia, Reykjavik'),
        ('W. Europe Standard Time', '(UTC+01:00) Amsterdam, Berlin, Bern, Rome, Stockholm, Vienna'),
        ('Central Europe Standard Time', '(UTC+01:00) Belgrade, Bratislava, Budapest, Ljubljana, Prague'),
        ('Romance Standard Time', '(UTC+01:00) Brussels, Copenhagen, Madrid, Paris'),
        ('Central European Standard Time', '(UTC+01:00) Sarajevo, Skopje, Warsaw, Zagreb'),
        ('W. Central Africa Standard Time', '(UTC+01:00) West Central Africa'),
        ('Namibia Standard Time', '(UTC+02:00) Windhoek'),
        ('Jordan Standard Time', '(UTC+02:00) Amman'),
        ('GTB Standard Time', '(UTC+02:00) Athens, Bucharest, Istanbul'),
        ('Middle East Standard Time', '(UTC+02:00) Beirut'),
        ('Egypt Standard Time', '(UTC+02:00) Cairo'),
        ('Syria Standard Time', '(UTC+02:00) Damascus'),
        ('South Africa Standard Time', '(UTC+02:00) Harare, Pretoria'),
        ('FLE Standard Time', '(UTC+02:00) Helsinki, Kyiv, Riga, Sofia, Tallinn, Vilnius'),
        ('Israel Standard Time', '(UTC+02:00) Jerusalem'),
        ('E. Europe Standard Time', '(UTC+02:00) Minsk'),
        ('Arabic Standard Time', '(UTC+03:00) Baghdad'),
        ('Arab Standard Time', '(UTC+03:00) Kuwait, Riyadh'),
        ('Russian Standard Time', '(UTC+03:00) Moscow, St. Petersburg, Volgograd'),
        ('E. Africa Standard Time', '(UTC+03:00) Nairobi'),
        ('Iran Standard Time', '(UTC+03:30) Tehran'),
        ('Georgian Standard Time', '(UTC+03:00) Tbilisi'),
        ('Arabian Standard Time', '(UTC+04:00) Abu Dhabi, Muscat'),
        ('Azerbaijan Standard Time', '(UTC+04:00) Baku'),
        ('Mauritius Standard Time', '(UTC+04:00) Port Louis'),
        ('Caucasus Standard Time', '(UTC+04:00) Yerevan'),
        ('Afghanistan Standard Time', '(UTC+04:30) Kabul'),
        ('Ekaterinburg Standard Time', '(UTC+05:00) Ekaterinburg'),
        ('Pakistan Standard Time', '(UTC+05:00) Islamabad, Karachi'),
        ('West Asia Standard Time', '(UTC+05:00) Tashkent'),
        ('India Standard Time', '(UTC+05:30) Chennai, Kolkata, Mumbai, New Delhi'),
        ('Sri Lanka Standard Time', '(UTC+05:30) Sri Jayawardenepura'),
        ('Nepal Standard Time', '(UTC+05:45) Kathmandu'),
        ('N. Central Asia Standard Time', '(UTC+06:00) Almaty, Novosibirsk'),
        ('Central Asia Standard Time', '(UTC+06:00) Astana'),
        ('Bangladesh Standard Time', '(UTC+06:00) Dhaka'),
        ('Myanmar Standard Time', '(UTC+06:30) Yangon (Rangoon)'),
        ('SE Asia Standard Time', '(UTC+07:00) Bangkok, Hanoi, Jakarta'),
        ('North Asia Standard Time', '(UTC+07:00) Krasnoyarsk'),
        ('China Standard Time', '(UTC+08:00) Beijing, Chongqing, Hong Kong, Urumqi'),
        ('North Asia East Standard Time', '(UTC+08:00) Irkutsk, Ulaan Bataar'),
        ('Singapore Standard Time', '(UTC+08:00) Kuala Lumpur, Singapore'),
        ('W. Australia Standard Time', '(UTC+08:00) Perth'),
        ('Taipei Standard Time', '(UTC+08:00) Taipei'),
        ('Ulaanbaatar Standard Time', '(UTC+08:00) Ulaanbaatar'),
        ('Tokyo Standard Time', '(UTC+09:00) Osaka, Sapporo, Tokyo'),
        ('Korea Standard Time', '(UTC+09:00) Seoul'),
        ('Yakutsk Standard Time', '(UTC+09:00) Yakutsk'),
        ('Cen. Australia Standard Time', '(UTC+09:30) Adelaide'),
        ('AUS Central Standard Time', '(UTC+09:30) Darwin'),
        ('E. Australia Standard Time', '(UTC+10:00) Brisbane'),
        ('AUS Eastern Standard Time', '(UTC+10:00) Canberra, Melbourne, Sydney'),
        ('West Pacific Standard Time', '(UTC+10:00) Guam, Port Moresby'),
        ('Tasmania Standard Time', '(UTC+10:00) Hobart'),
        ('Vladivostok Standard Time', '(UTC+10:00) Vladivostok'),
        ('Magadan Standard Time', '(UTC+11:00) Magadan'),
        ('Central Pacific Standard Time', '(UTC+11:00) Magadan, Solomon Is., New Caledonia'),
        ('New Zealand Standard Time', '(UTC+12:00) Auckland, Wellington'),
        ('UTC+12', '(UTC+12:00) Coordinated Universal Time+12'),
        ('Fiji Standard Time', '(UTC+12:00) Fiji, Marshall Is.'),
        ('Kamchatka Standard Time', '(UTC+12:00) Petropavlovsk-Kamchatsky'),
        ('Tonga Standard Time', "(UTC+13:00) Nuku'alofa"),
    ])



import logging

from .credentials import Credentials
from .ewsdatetime import EWSTimeZone
from .protocol import Protocol

log = logging.getLogger(__name__)


class Configuration:
    """
    Stores default credentials when connecting via a system account, and default connection protocol when autodiscover
    is not activated for an Account.
    """
    def __init__(self, server=None, username=None, password=None, has_ssl=True, ews_auth_type=None, ews_url=None):
        if username:
            if not password:
                raise AttributeError('Password must be provided when username is provided')
            self.credentials = Credentials(username, password)
        else:
            self.credentials = None
        if server or ews_url:
            if not self.credentials:
                raise AttributeError('Credentials must be provided when server is provided')
            # Set up a default protocol that non-autodiscover accounts can use
            if not ews_url:
                ews_url = '%s://%s/EWS/Exchange.asmx' % ('https' if has_ssl else 'http', server)
            self.protocol = Protocol(
                ews_url=ews_url,
                ews_auth_type=ews_auth_type,
                has_ssl=has_ssl,
                credentials=self.credentials,
            )
        else:
            self.protocol = None



"""
A protocol is an endpoint for EWS service connections. It contains all necessary information to make HTTPS connections.

Protocols should be accessed through an Account, and are either created from a default Configuration or autodiscovered
when creating an Account.
"""
import socket
import queue
from multiprocessing.pool import ThreadPool
import logging
from collections import defaultdict
from threading import Lock
from urllib import parse
import random

from requests import adapters, Session

from .credentials import Credentials
from .errors import TransportError
from .ewsdatetime import EWSTimeZone
from .transport import get_auth_instance, get_service_authtype, get_docs_authtype, test_credentials
from .version import Version, API_VERSIONS

log = logging.getLogger(__name__)

# Used to cache version and auth types for visited servers
_server_cache = defaultdict(dict)
_server_cache_lock = Lock()

POOLSIZE = 4
TIMEOUT = 120


def close_connections():
    for server, cached_values in _server_cache.items():
        cached_protocol = Protocol('https://%s/EWS/Exchange.asmx' % server, True, Credentials('', ''))
        cached_protocol.close()


class Protocol:
    SESSION_POOLSIZE = 1

    def __init__(self, ews_url, has_ssl, credentials, ews_auth_type=None, timezone='GMT'):
        assert isinstance(credentials, Credentials)
        self.server = parse.urlparse(ews_url).hostname.lower()
        self.has_ssl = has_ssl
        self.ews_url = ews_url
        scheme = 'https' if self.has_ssl else 'http'
        self.wsdl_url = '%s://%s/EWS/Services.wsdl' % (scheme, self.server)
        self.messages_url = '%s://%s/EWS/messages.xsd' % (scheme, self.server)
        self.types_url = '%s://%s/EWS/types.xsd' % (scheme, self.server)
        self.credentials = credentials
        self.timeout = TIMEOUT

        # Acquire lock to guard against multiple threads competing to cache information. Having a per-server lock is
        # overkill.
        log.debug('Waiting for _server_cache_lock')
        with _server_cache_lock:
            if self.server in _server_cache:
                # Get cached version and auth types and session / thread pools
                log.debug("Cache hit for server '%s'", self.server)
                for k, v in _server_cache[self.server].items():
                    setattr(self, k, v)

                if ews_auth_type:
                    if ews_auth_type != self.ews_auth_type:
                        # Some Exchange servers just can't make up their mind
                        log.debug('Auth type mismatch on server %s. %s != %s' % (self.server, ews_auth_type,
                                                                                 self.ews_auth_type))
            else:
                log.debug("Cache miss. Adding server '%s', poolsize %s, timeout %s", self.server, POOLSIZE, TIMEOUT)
                # Autodetect authentication type if necessary
                self.ews_auth_type = ews_auth_type or get_service_authtype(server=self.server, has_ssl=self.has_ssl,
                                                                           ews_url=ews_url, versions=API_VERSIONS)
                self.docs_auth_type = get_docs_authtype(server=self.server, has_ssl=self.has_ssl, url=self.types_url)

                # Try to behave nicely with the Exchange server. We want to keep the connection open between requests.
                # We also want to re-use sessions, to avoid the NTLM auth handshake on every request.
                assert POOLSIZE > 0
                self._session_pool = queue.LifoQueue(maxsize=POOLSIZE)
                for i in range(POOLSIZE):
                    self._session_pool.put(self.create_session(), block=False)

                # Used by services to process service requests that are able to run in parallel. Thread pool should be
                # larger than connection the pool so we have time to process data without idling the connection.
                thread_poolsize = 4 * POOLSIZE
                self.thread_pool = ThreadPool(processes=thread_poolsize)

                # Needs auth objects and a working session pool
                self.version = Version.guess(self)

                # Cache results
                _server_cache[self.server] = dict(
                    version=self.version,
                    ews_auth_type=self.ews_auth_type,
                    docs_auth_type=self.docs_auth_type,
                    thread_pool=self.thread_pool,
                    _session_pool=self._session_pool,
                )
        log.debug('_server_cache_lock released')

    def close(self):
        log.debug('Server %s: Closing sessions', self.server)
        while True:
            try:
                self._session_pool.get(block=False).close_socket(self.ews_url)
            except queue.Empty:
                break

    def get_session(self):
        _timeout = 60  # Rate-limit messages about session starvation
        while True:
            try:
                log.debug('Server %s: Waiting for session', self.server)
                session = self._session_pool.get(timeout=_timeout)
                log.debug('Server %s: Got session %s', self.server, session.session_id)
                return session
            except queue.Empty:
                # This is normal when we have many worker threads starving for available sessions
                log.debug('Server %s: No sessions available for %s seconds', self.server, _timeout)

    def release_session(self, session):
        # This should never fail, as we don't have more sessions than the queue contains
        log.debug('Server %s: Releasing session %s', self.server, session.session_id)
        try:
            self._session_pool.put(session, block=False)
        except queue.Full:
            log.debug('Server %s: Session pool was already full %s', self.server, session.session_id)

    def retire_session(self, session):
        # The session is useless. Close it completely and place a fresh session in the pool
        log.debug('Server %s: Retiring session %s', self.server, session.session_id)
        session.close_socket(self.ews_url)
        del session
        self.release_session(self.create_session())

    def renew_session(self, session):
        # The session is useless. Close it completely and place a fresh session in the pool
        log.debug('Server %s: Renewing session %s', self.server, session.session_id)
        session.close_socket(self.ews_url)
        del session
        return self.create_session()

    def test(self):
        # We need the version for this
        try:
            socket.gethostbyname_ex(self.server)[2][0]
        except socket.gaierror as e:
            raise TransportError("Server '%s' does not exist" % self.server) from e
        return test_credentials(protocol=self)

    def __str__(self):
        return '''\
EWS url: %s
Product version (according to XSD): %s
API version (according to SOAP headers): %s
Full name: %s
Build numbers: %s
EWS auth: %s
XSD auth: %s''' % (
            self.ews_url,
            self.version.shortname,
            self.version.api_version,
            self.version.name,
            self.version.build,
            self.ews_auth_type,
            self.docs_auth_type,
        )

    def create_session(self):
        session = EWSSession(self)
        session.auth = get_auth_instance(credentials=self.credentials, auth_type=self.ews_auth_type)
        # Leave this inside the loop because headers are mutable
        headers = {'Content-Type': 'text/xml; charset=utf-8', 'Accept-Encoding': 'compress, gzip'}
        session.headers.update(headers)
        scheme = 'https' if self.has_ssl else 'http'
        # We want just one connection per session. No retries, since we wrap all requests in our own retry handler
        assert self.SESSION_POOLSIZE > 0
        session.mount('%s://' % scheme, adapters.HTTPAdapter(pool_block=True, pool_connections=self.SESSION_POOLSIZE,
                                                             pool_maxsize=self.SESSION_POOLSIZE, max_retries=0))
        log.debug('Server %s: Created session %s', self.server, session.session_id)
        return session


class EWSSession(Session):
    def __init__(self, protocol):
        self.session_id = random.randint(1, 32767)  # Used for debugging messages in services
        self.protocol = protocol
        super().__init__()

    def close_socket(self, url):
        # Close underlying socket. This ensures we don't leave stray sockets around after program exit.
        adapter = self.get_adapter(url)
        pool = adapter.get_connection(url)
        for i in range(pool.pool.qsize()):
            conn = pool._get_conn()
            if conn.sock:
                log.debug('Closing socket %s', str(conn.sock.getsockname()))
                conn.sock.shutdown(socket.SHUT_RDWR)
                conn.sock.close()

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.protocol.release_session(self)
        else:
            self.protocol.retire_session(self)
        # return super().__exit__()  # We don't want to close the session socket when we have used the session



"""
Stores errors specific to exchangelib, and mirrors all the possible errors that EWS can return.
"""


class EWSError(Exception):
    """
    Global error type within this module.

    """
    def __init__(self, value):
        super().__init__(value)
        self.value = value

    def __str__(self):
        return str(self.value)


# Warnings
class EWSWarning(EWSError):
    pass


# Misc errors
class TransportError(EWSError):
    pass


class RateLimitError(TransportError):
    pass


class SOAPError(TransportError):
    pass


class UnauthorizedError(EWSError):
    pass


class RedirectError(TransportError):
    def __init__(self, url, server, has_ssl):
        self.url = url
        self.server = server
        self.has_ssl = has_ssl
        super().__init__(str(self))

    def __str__(self):
        return 'We were redirected to %s, server %s, has_ssl %s' % (self.url, self.server, self.has_ssl)


class AutoDiscoverError(TransportError):
    pass


class AutoDiscoverFailed(AutoDiscoverError):
    pass


class AutoDiscoverRedirect(AutoDiscoverError):
    def __init__(self, redirect_email):
        self.redirect_email = redirect_email
        super().__init__(str(self))

    def __str__(self):
        return 'AutoDiscover redirects to %s' % self.redirect_email


class ResponseMessageError(TransportError):
    pass


# Somewhat-authoritative list of possible response message error types from EWS. See full list at
# https://msdn.microsoft.com/en-us/library/office/aa580757(v=exchg.150).aspx
#
class ErrorAccessDenied(ResponseMessageError): pass
class ErrorAccessModeSpecified(ResponseMessageError): pass
class ErrorAccountDisabled(ResponseMessageError): pass
class ErrorAddDelegatesFailed(ResponseMessageError): pass
class ErrorAddressSpaceNotFound(ResponseMessageError): pass
class ErrorADOperation(ResponseMessageError): pass
class ErrorADSessionFilter(ResponseMessageError): pass
class ErrorADUnavailable(ResponseMessageError): pass
class ErrorAffectedTaskOccurrencesRequired(ResponseMessageError): pass
class ErrorApplyConversationActionFailed(ResponseMessageError): pass
class ErrorAttachmentSizeLimitExceeded(ResponseMessageError): pass
class ErrorAutoDiscoverFailed(ResponseMessageError): pass
class ErrorAvailabilityConfigNotFound(ResponseMessageError): pass
class ErrorBatchProcessingStopped(ResponseMessageError): pass
class ErrorCalendarCannotMoveOrCopyOccurrence(ResponseMessageError): pass
class ErrorCalendarCannotUpdateDeletedItem(ResponseMessageError): pass
class ErrorCalendarCannotUseIdForOccurrenceId(ResponseMessageError): pass
class ErrorCalendarCannotUseIdForRecurringMasterId(ResponseMessageError): pass
class ErrorCalendarDurationIsTooLong(ResponseMessageError): pass
class ErrorCalendarEndDateIsEarlierThanStartDate(ResponseMessageError): pass
class ErrorCalendarFolderIsInvalidForCalendarView(ResponseMessageError): pass
class ErrorCalendarInvalidAttributeValue(ResponseMessageError): pass
class ErrorCalendarInvalidDayForTimeChangePattern(ResponseMessageError): pass
class ErrorCalendarInvalidDayForWeeklyRecurrence(ResponseMessageError): pass
class ErrorCalendarInvalidPropertyState(ResponseMessageError): pass
class ErrorCalendarInvalidPropertyValue(ResponseMessageError): pass
class ErrorCalendarInvalidRecurrence(ResponseMessageError): pass
class ErrorCalendarInvalidTimeZone(ResponseMessageError): pass
class ErrorCalendarIsCancelledForAccept(ResponseMessageError): pass
class ErrorCalendarIsCancelledForDecline(ResponseMessageError): pass
class ErrorCalendarIsCancelledForRemove(ResponseMessageError): pass
class ErrorCalendarIsCancelledForTentative(ResponseMessageError): pass
class ErrorCalendarIsDelegatedForAccept(ResponseMessageError): pass
class ErrorCalendarIsDelegatedForDecline(ResponseMessageError): pass
class ErrorCalendarIsDelegatedForRemove(ResponseMessageError): pass
class ErrorCalendarIsDelegatedForTentative(ResponseMessageError): pass
class ErrorCalendarIsNotOrganizer(ResponseMessageError): pass
class ErrorCalendarIsOrganizerForAccept(ResponseMessageError): pass
class ErrorCalendarIsOrganizerForDecline(ResponseMessageError): pass
class ErrorCalendarIsOrganizerForRemove(ResponseMessageError): pass
class ErrorCalendarIsOrganizerForTentative(ResponseMessageError): pass
class ErrorCalendarMeetingRequestIsOutOfDate(ResponseMessageError): pass
class ErrorCalendarOccurrenceIndexIsOutOfRecurrenceRange(ResponseMessageError): pass
class ErrorCalendarOccurrenceIsDeletedFromRecurrence(ResponseMessageError): pass
class ErrorCalendarOutOfRange(ResponseMessageError): pass
class ErrorCalendarViewRangeTooBig(ResponseMessageError): pass
class ErrorCallerIsInvalidADAccount(ResponseMessageError): pass
class ErrorCannotCreateCalendarItemInNonCalendarFolder(ResponseMessageError): pass
class ErrorCannotCreateContactInNonContactFolder(ResponseMessageError): pass
class ErrorCannotCreatePostItemInNonMailFolder(ResponseMessageError): pass
class ErrorCannotCreateTaskInNonTaskFolder(ResponseMessageError): pass
class ErrorCannotDeleteObject(ResponseMessageError): pass
class ErrorCannotDeleteTaskOccurrence(ResponseMessageError): pass
class ErrorCannotEmptyFolder(ResponseMessageError): pass
class ErrorCannotOpenFileAttachment(ResponseMessageError): pass
class ErrorCannotSetCalendarPermissionOnNonCalendarFolder(ResponseMessageError): pass
class ErrorCannotSetNonCalendarPermissionOnCalendarFolder(ResponseMessageError): pass
class ErrorCannotSetPermissionUnknownEntries(ResponseMessageError): pass
class ErrorCannotUseFolderIdForItemId(ResponseMessageError): pass
class ErrorCannotUseItemIdForFolderId(ResponseMessageError): pass
class ErrorChangeKeyRequired(ResponseMessageError): pass
class ErrorChangeKeyRequiredForWriteOperations(ResponseMessageError): pass
class ErrorClientDisconnected(ResponseMessageError): pass
class ErrorConnectionFailed(ResponseMessageError): pass
class ErrorContainsFilterWrongType(ResponseMessageError): pass
class ErrorContentConversionFailed(ResponseMessageError): pass
class ErrorCorruptData(ResponseMessageError): pass
class ErrorCreateItemAccessDenied(ResponseMessageError): pass
class ErrorCreateManagedFolderPartialCompletion(ResponseMessageError): pass
class ErrorCreateSubfolderAccessDenied(ResponseMessageError): pass
class ErrorCrossMailboxMoveCopy(ResponseMessageError): pass
class ErrorCrossSiteRequest(ResponseMessageError): pass
class ErrorDataSizeLimitExceeded(ResponseMessageError): pass
class ErrorDataSourceOperation(ResponseMessageError): pass
class ErrorDelegateAlreadyExists(ResponseMessageError): pass
class ErrorDelegateCannotAddOwner(ResponseMessageError): pass
class ErrorDelegateMissingConfiguration(ResponseMessageError): pass
class ErrorDelegateNoUser(ResponseMessageError): pass
class ErrorDelegateValidationFailed(ResponseMessageError): pass
class ErrorDeleteDistinguishedFolder(ResponseMessageError): pass
class ErrorDeleteItemsFailed(ResponseMessageError): pass
class ErrorDistinguishedUserNotSupported(ResponseMessageError): pass
class ErrorDistributionListMemberNotExist(ResponseMessageError): pass
class ErrorDuplicateInputFolderNames(ResponseMessageError): pass
class ErrorDuplicateSOAPHeader(ResponseMessageError): pass
class ErrorDuplicateUserIdsSpecified(ResponseMessageError): pass
class ErrorEmailAddressMismatch(ResponseMessageError): pass
class ErrorEventNotFound(ResponseMessageError): pass
class ErrorExceededConnectionCount(ResponseMessageError): pass
class ErrorExceededFindCountLimit(ResponseMessageError): pass
class ErrorExceededSubscriptionCount(ResponseMessageError): pass
class ErrorExpiredSubscription(ResponseMessageError): pass
class ErrorFolderCorrupt(ResponseMessageError): pass
class ErrorFolderExists(ResponseMessageError): pass
class ErrorFolderNotFound(ResponseMessageError): pass
class ErrorFolderPropertyRequestFailed(ResponseMessageError): pass
class ErrorFolderSave(ResponseMessageError): pass
class ErrorFolderSaveFailed(ResponseMessageError): pass
class ErrorFolderSavePropertyError(ResponseMessageError): pass
class ErrorFreeBusyDLLimitReached(ResponseMessageError): pass
class ErrorFreeBusyGenerationFailed(ResponseMessageError): pass
class ErrorGetServerSecurityDescriptorFailed(ResponseMessageError): pass
class ErrorImpersonateUserDenied(ResponseMessageError): pass
class ErrorImpersonationDenied(ResponseMessageError): pass
class ErrorImpersonationFailed(ResponseMessageError): pass
class ErrorInboxRulesValidationError(ResponseMessageError): pass
class ErrorIncorrectSchemaVersion(ResponseMessageError): pass
class ErrorIncorrectUpdatePropertyCount(ResponseMessageError): pass
class ErrorIndividualMailboxLimitReached(ResponseMessageError): pass
class ErrorInsufficientResources(ResponseMessageError): pass
class ErrorInternalServerError(ResponseMessageError): pass
class ErrorInternalServerTransientError(ResponseMessageError): pass
class ErrorInvalidAccessLevel(ResponseMessageError): pass
class ErrorInvalidArgument(ResponseMessageError): pass
class ErrorInvalidAttachmentId(ResponseMessageError): pass
class ErrorInvalidAttachmentSubfilter(ResponseMessageError): pass
class ErrorInvalidAttachmentSubfilterTextFilter(ResponseMessageError): pass
class ErrorInvalidAuthorizationContext(ResponseMessageError): pass
class ErrorInvalidChangeKey(ResponseMessageError): pass
class ErrorInvalidClientSecurityContext(ResponseMessageError): pass
class ErrorInvalidCompleteDate(ResponseMessageError): pass
class ErrorInvalidContactEmailAddress(ResponseMessageError): pass
class ErrorInvalidContactEmailIndex(ResponseMessageError): pass
class ErrorInvalidCrossForestCredentials(ResponseMessageError): pass
class ErrorInvalidDelegatePermission(ResponseMessageError): pass
class ErrorInvalidDelegateUserId(ResponseMessageError): pass
class ErrorInvalidExchangeImpersonationHeaderData(ResponseMessageError): pass
class ErrorInvalidExcludesRestriction(ResponseMessageError): pass
class ErrorInvalidExpressionTypeForSubFilter(ResponseMessageError): pass
class ErrorInvalidExtendedProperty(ResponseMessageError): pass
class ErrorInvalidExtendedPropertyValue(ResponseMessageError): pass
class ErrorInvalidExternalSharingInitiator(ResponseMessageError): pass
class ErrorInvalidExternalSharingSubscriber(ResponseMessageError): pass
class ErrorInvalidFederatedOrganizationId(ResponseMessageError): pass
class ErrorInvalidFolderId(ResponseMessageError): pass
class ErrorInvalidFolderTypeForOperation(ResponseMessageError): pass
class ErrorInvalidFractionalPagingParameters(ResponseMessageError): pass
class ErrorInvalidFreeBusyViewType(ResponseMessageError): pass
class ErrorInvalidGetSharingFolderRequest(ResponseMessageError): pass
class ErrorInvalidId(ResponseMessageError): pass
class ErrorInvalidIdEmpty(ResponseMessageError): pass
class ErrorInvalidIdMalformed(ResponseMessageError): pass
class ErrorInvalidIdMalformedEwsLegacyIdFormat(ResponseMessageError): pass
class ErrorInvalidIdMonikerTooLong(ResponseMessageError): pass
class ErrorInvalidIdNotAnItemAttachmentId(ResponseMessageError): pass
class ErrorInvalidIdReturnedByResolveNames(ResponseMessageError): pass
class ErrorInvalidIdStoreObjectIdTooLong(ResponseMessageError): pass
class ErrorInvalidIdTooManyAttachmentLevels(ResponseMessageError): pass
class ErrorInvalidIdXml(ResponseMessageError): pass
class ErrorInvalidIndexedPagingParameters(ResponseMessageError): pass
class ErrorInvalidInternetHeaderChildNodes(ResponseMessageError): pass
class ErrorInvalidItemForOperationAcceptItem(ResponseMessageError): pass
class ErrorInvalidItemForOperationCancelItem(ResponseMessageError): pass
class ErrorInvalidItemForOperationCreateItem(ResponseMessageError): pass
class ErrorInvalidItemForOperationCreateItemAttachment(ResponseMessageError): pass
class ErrorInvalidItemForOperationDeclineItem(ResponseMessageError): pass
class ErrorInvalidItemForOperationExpandDL(ResponseMessageError): pass
class ErrorInvalidItemForOperationRemoveItem(ResponseMessageError): pass
class ErrorInvalidItemForOperationSendItem(ResponseMessageError): pass
class ErrorInvalidItemForOperationTentative(ResponseMessageError): pass
class ErrorInvalidLicense(ResponseMessageError): pass
class ErrorInvalidLogonType(ResponseMessageError): pass
class ErrorInvalidMailbox(ResponseMessageError): pass
class ErrorInvalidManagedFolderProperty(ResponseMessageError): pass
class ErrorInvalidManagedFolderQuota(ResponseMessageError): pass
class ErrorInvalidManagedFolderSize(ResponseMessageError): pass
class ErrorInvalidMergedFreeBusyInterval(ResponseMessageError): pass
class ErrorInvalidNameForNameResolution(ResponseMessageError): pass
class ErrorInvalidNetworkServiceContext(ResponseMessageError): pass
class ErrorInvalidOofParameter(ResponseMessageError): pass
class ErrorInvalidOperation(ResponseMessageError): pass
class ErrorInvalidOrganizationRelationshipForFreeBusy(ResponseMessageError): pass
class ErrorInvalidPagingMaxRows(ResponseMessageError): pass
class ErrorInvalidParentFolder(ResponseMessageError): pass
class ErrorInvalidPercentCompleteValue(ResponseMessageError): pass
class ErrorInvalidPermissionSettings(ResponseMessageError): pass
class ErrorInvalidPhoneCallId(ResponseMessageError): pass
class ErrorInvalidPhoneNumber(ResponseMessageError): pass
class ErrorInvalidPropertyAppend(ResponseMessageError): pass
class ErrorInvalidPropertyDelete(ResponseMessageError): pass
class ErrorInvalidPropertyForExists(ResponseMessageError): pass
class ErrorInvalidPropertyForOperation(ResponseMessageError): pass
class ErrorInvalidPropertyRequest(ResponseMessageError): pass
class ErrorInvalidPropertySet(ResponseMessageError): pass
class ErrorInvalidPropertyUpdateSentMessage(ResponseMessageError): pass
class ErrorInvalidProxySecurityContext(ResponseMessageError): pass
class ErrorInvalidPullSubscriptionId(ResponseMessageError): pass
class ErrorInvalidPushSubscriptionUrl(ResponseMessageError): pass
class ErrorInvalidRecipients(ResponseMessageError): pass
class ErrorInvalidRecipientSubfilter(ResponseMessageError): pass
class ErrorInvalidRecipientSubfilterComparison(ResponseMessageError): pass
class ErrorInvalidRecipientSubfilterOrder(ResponseMessageError): pass
class ErrorInvalidRecipientSubfilterTextFilter(ResponseMessageError): pass
class ErrorInvalidReferenceItem(ResponseMessageError): pass
class ErrorInvalidRequest(ResponseMessageError): pass
class ErrorInvalidRestriction(ResponseMessageError): pass
class ErrorInvalidRoutingType(ResponseMessageError): pass
class ErrorInvalidScheduledOofDuration(ResponseMessageError): pass
class ErrorInvalidSchemaVersionForMailboxVersion(ResponseMessageError): pass
class ErrorInvalidSecurityDescriptor(ResponseMessageError): pass
class ErrorInvalidSendItemSaveSettings(ResponseMessageError): pass
class ErrorInvalidSerializedAccessToken(ResponseMessageError): pass
class ErrorInvalidServerVersion(ResponseMessageError): pass
class ErrorInvalidSharingData(ResponseMessageError): pass
class ErrorInvalidSharingMessage(ResponseMessageError): pass
class ErrorInvalidSid(ResponseMessageError): pass
class ErrorInvalidSIPUri(ResponseMessageError): pass
class ErrorInvalidSmtpAddress(ResponseMessageError): pass
class ErrorInvalidSubfilterType(ResponseMessageError): pass
class ErrorInvalidSubfilterTypeNotAttendeeType(ResponseMessageError): pass
class ErrorInvalidSubfilterTypeNotRecipientType(ResponseMessageError): pass
class ErrorInvalidSubscription(ResponseMessageError): pass
class ErrorInvalidSubscriptionRequest(ResponseMessageError): pass
class ErrorInvalidSyncStateData(ResponseMessageError): pass
class ErrorInvalidTimeInterval(ResponseMessageError): pass
class ErrorInvalidUserInfo(ResponseMessageError): pass
class ErrorInvalidUserOofSettings(ResponseMessageError): pass
class ErrorInvalidUserPrincipalName(ResponseMessageError): pass
class ErrorInvalidUserSid(ResponseMessageError): pass
class ErrorInvalidUserSidMissingUPN(ResponseMessageError): pass
class ErrorInvalidValueForProperty(ResponseMessageError): pass
class ErrorInvalidWatermark(ResponseMessageError): pass
class ErrorIPGatewayNotFound(ResponseMessageError): pass
class ErrorIrresolvableConflict(ResponseMessageError): pass
class ErrorItemCorrupt(ResponseMessageError): pass
class ErrorItemNotFound(ResponseMessageError): pass
class ErrorItemPropertyRequestFailed(ResponseMessageError): pass
class ErrorItemSave(ResponseMessageError): pass
class ErrorItemSavePropertyError(ResponseMessageError): pass
class ErrorLegacyMailboxFreeBusyViewTypeNotMerged(ResponseMessageError): pass
class ErrorLocalServerObjectNotFound(ResponseMessageError): pass
class ErrorLogonAsNetworkServiceFailed(ResponseMessageError): pass
class ErrorMailboxConfiguration(ResponseMessageError): pass
class ErrorMailboxDataArrayEmpty(ResponseMessageError): pass
class ErrorMailboxDataArrayTooBig(ResponseMessageError): pass
class ErrorMailboxFailover(ResponseMessageError): pass
class ErrorMailboxLogonFailed(ResponseMessageError): pass
class ErrorMailboxMoveInProgress(ResponseMessageError): pass
class ErrorMailboxStoreUnavailable(ResponseMessageError): pass
class ErrorMailRecipientNotFound(ResponseMessageError): pass
class ErrorMailTipsDisabled(ResponseMessageError): pass
class ErrorManagedFolderAlreadyExists(ResponseMessageError): pass
class ErrorManagedFolderNotFound(ResponseMessageError): pass
class ErrorManagedFoldersRootFailure(ResponseMessageError): pass
class ErrorMeetingSuggestionGenerationFailed(ResponseMessageError): pass
class ErrorMessageDispositionRequired(ResponseMessageError): pass
class ErrorMessageSizeExceeded(ResponseMessageError): pass
class ErrorMessageTrackingNoSuchDomain(ResponseMessageError): pass
class ErrorMessageTrackingPermanentError(ResponseMessageError): pass
class ErrorMessageTrackingTransientError(ResponseMessageError): pass
class ErrorMimeContentConversionFailed(ResponseMessageError): pass
class ErrorMimeContentInvalid(ResponseMessageError): pass
class ErrorMimeContentInvalidBase64String(ResponseMessageError): pass
class ErrorMissedNotificationEvents(ResponseMessageError): pass
class ErrorMissingArgument(ResponseMessageError): pass
class ErrorMissingEmailAddress(ResponseMessageError): pass
class ErrorMissingEmailAddressForManagedFolder(ResponseMessageError): pass
class ErrorMissingInformationEmailAddress(ResponseMessageError): pass
class ErrorMissingInformationReferenceItemId(ResponseMessageError): pass
class ErrorMissingInformationSharingFolderId(ResponseMessageError): pass
class ErrorMissingItemForCreateItemAttachment(ResponseMessageError): pass
class ErrorMissingManagedFolderId(ResponseMessageError): pass
class ErrorMissingRecipients(ResponseMessageError): pass
class ErrorMissingUserIdInformation(ResponseMessageError): pass
class ErrorMoreThanOneAccessModeSpecified(ResponseMessageError): pass
class ErrorMoveCopyFailed(ResponseMessageError): pass
class ErrorMoveDistinguishedFolder(ResponseMessageError): pass
class ErrorNameResolutionMultipleResults(ResponseMessageError): pass
class ErrorNameResolutionNoMailbox(ResponseMessageError): pass
class ErrorNameResolutionNoResults(ResponseMessageError): pass
class ErrorNewEventStreamConnectionOpened(ResponseMessageError): pass
class ErrorNoApplicableProxyCASServersAvailable(ResponseMessageError): pass
class ErrorNoCalendar(ResponseMessageError): pass
class ErrorNoDestinationCASDueToKerberosRequirements(ResponseMessageError): pass
class ErrorNoDestinationCASDueToSSLRequirements(ResponseMessageError): pass
class ErrorNoDestinationCASDueToVersionMismatch(ResponseMessageError): pass
class ErrorNoFolderClassOverride(ResponseMessageError): pass
class ErrorNoFreeBusyAccess(ResponseMessageError): pass
class ErrorNonExistentMailbox(ResponseMessageError): pass
class ErrorNonPrimarySmtpAddress(ResponseMessageError): pass
class ErrorNoPropertyTagForCustomProperties(ResponseMessageError): pass
class ErrorNoPublicFolderReplicaAvailable(ResponseMessageError): pass
class ErrorNoPublicFolderServerAvailable(ResponseMessageError): pass
class ErrorNoRespondingCASInDestinationSite(ResponseMessageError): pass
class ErrorNotAllowedExternalSharingByPolicy(ResponseMessageError): pass
class ErrorNotDelegate(ResponseMessageError): pass
class ErrorNotEnoughMemory(ResponseMessageError): pass
class ErrorNotSupportedSharingMessage(ResponseMessageError): pass
class ErrorObjectTypeChanged(ResponseMessageError): pass
class ErrorOccurrenceCrossingBoundary(ResponseMessageError): pass
class ErrorOccurrenceTimeSpanTooBig(ResponseMessageError): pass
class ErrorOperationNotAllowedWithPublicFolderRoot(ResponseMessageError): pass
class ErrorOrganizationNotFederated(ResponseMessageError): pass
class ErrorOutlookRuleBlobExists(ResponseMessageError): pass
class ErrorParentFolderIdRequired(ResponseMessageError): pass
class ErrorParentFolderNotFound(ResponseMessageError): pass
class ErrorPasswordChangeRequired(ResponseMessageError): pass
class ErrorPasswordExpired(ResponseMessageError): pass
class ErrorPermissionNotAllowedByPolicy(ResponseMessageError): pass
class ErrorPhoneNumberNotDialable(ResponseMessageError): pass
class ErrorPropertyUpdate(ResponseMessageError): pass
class ErrorPropertyValidationFailure(ResponseMessageError): pass
class ErrorProxiedSubscriptionCallFailure(ResponseMessageError): pass
class ErrorProxyCallFailed(ResponseMessageError): pass
class ErrorProxyGroupSidLimitExceeded(ResponseMessageError): pass
class ErrorProxyRequestNotAllowed(ResponseMessageError): pass
class ErrorProxyRequestProcessingFailed(ResponseMessageError): pass
class ErrorProxyServiceDiscoveryFailed(ResponseMessageError): pass
class ErrorProxyTokenExpired(ResponseMessageError): pass
class ErrorPublicFolderRequestProcessingFailed(ResponseMessageError): pass
class ErrorPublicFolderServerNotFound(ResponseMessageError): pass
class ErrorQueryFilterTooLong(ResponseMessageError): pass
class ErrorQuotaExceeded(ResponseMessageError): pass
class ErrorReadEventsFailed(ResponseMessageError): pass
class ErrorReadReceiptNotPending(ResponseMessageError): pass
class ErrorRecurrenceEndDateTooBig(ResponseMessageError): pass
class ErrorRecurrenceHasNoOccurrence(ResponseMessageError): pass
class ErrorRemoveDelegatesFailed(ResponseMessageError): pass
class ErrorRequestAborted(ResponseMessageError): pass
class ErrorRequestStreamTooBig(ResponseMessageError): pass
class ErrorRequiredPropertyMissing(ResponseMessageError): pass
class ErrorResolveNamesInvalidFolderType(ResponseMessageError): pass
class ErrorResolveNamesOnlyOneContactsFolderAllowed(ResponseMessageError): pass
class ErrorResponseSchemaValidation(ResponseMessageError): pass
class ErrorRestrictionTooComplex(ResponseMessageError): pass
class ErrorRestrictionTooLong(ResponseMessageError): pass
class ErrorResultSetTooBig(ResponseMessageError): pass
class ErrorRulesOverQuota(ResponseMessageError): pass
class ErrorSavedItemFolderNotFound(ResponseMessageError): pass
class ErrorSchemaValidation(ResponseMessageError): pass
class ErrorSearchFolderNotInitialized(ResponseMessageError): pass
class ErrorSendAsDenied(ResponseMessageError): pass
class ErrorSendMeetingCancellationsRequired(ResponseMessageError): pass
class ErrorSendMeetingInvitationsOrCancellationsRequired(ResponseMessageError): pass
class ErrorSendMeetingInvitationsRequired(ResponseMessageError): pass
class ErrorSentMeetingRequestUpdate(ResponseMessageError): pass
class ErrorSentTaskRequestUpdate(ResponseMessageError): pass
class ErrorServerBusy(ResponseMessageError): pass
class ErrorServiceDiscoveryFailed(ResponseMessageError): pass
class ErrorSharingNoExternalEwsAvailable(ResponseMessageError): pass
class ErrorSharingSynchronizationFailed(ResponseMessageError): pass
class ErrorStaleObject(ResponseMessageError): pass
class ErrorSubmissionQuotaExceeded(ResponseMessageError): pass
class ErrorSubscriptionAccessDenied(ResponseMessageError): pass
class ErrorSubscriptionDelegateAccessNotSupported(ResponseMessageError): pass
class ErrorSubscriptionNotFound(ResponseMessageError): pass
class ErrorSubscriptionUnsubsribed(ResponseMessageError): pass
class ErrorSyncFolderNotFound(ResponseMessageError): pass
class ErrorTimeIntervalTooBig(ResponseMessageError): pass
class ErrorTimeoutExpired(ResponseMessageError): pass
class ErrorTimeZone(ResponseMessageError): pass
class ErrorToFolderNotFound(ResponseMessageError): pass
class ErrorTokenSerializationDenied(ResponseMessageError): pass
class ErrorTooManyObjectsOpened(ResponseMessageError): pass
class ErrorUnableToGetUserOofSettings(ResponseMessageError): pass
class ErrorUnifiedMessagingDialPlanNotFound(ResponseMessageError): pass
class ErrorUnifiedMessagingRequestFailed(ResponseMessageError): pass
class ErrorUnifiedMessagingServerNotFound(ResponseMessageError): pass
class ErrorUnsupportedCulture(ResponseMessageError): pass
class ErrorUnsupportedMapiPropertyType(ResponseMessageError): pass
class ErrorUnsupportedMimeConversion(ResponseMessageError): pass
class ErrorUnsupportedPathForQuery(ResponseMessageError): pass
class ErrorUnsupportedPathForSortGroup(ResponseMessageError): pass
class ErrorUnsupportedPropertyDefinition(ResponseMessageError): pass
class ErrorUnsupportedQueryFilter(ResponseMessageError): pass
class ErrorUnsupportedRecurrence(ResponseMessageError): pass
class ErrorUnsupportedSubFilter(ResponseMessageError): pass
class ErrorUnsupportedTypeForConversion(ResponseMessageError): pass
class ErrorUpdateDelegatesFailed(ResponseMessageError): pass
class ErrorUpdatePropertyMismatch(ResponseMessageError): pass
class ErrorUserNotAllowedByPolicy(ResponseMessageError): pass
class ErrorUserNotUnifiedMessagingEnabled(ResponseMessageError): pass
class ErrorUserWithoutFederatedProxyAddress(ResponseMessageError): pass
class ErrorValueOutOfRange(ResponseMessageError): pass
class ErrorVirusDetected(ResponseMessageError): pass
class ErrorVirusMessageDeleted(ResponseMessageError): pass
class ErrorVoiceMailNotImplemented(ResponseMessageError): pass
class ErrorWebRequestInInvalidState(ResponseMessageError): pass
class ErrorWin32InteropError(ResponseMessageError): pass
class ErrorWorkingHoursSaveFailed(ResponseMessageError): pass
class ErrorWorkingHoursXmlMalformed(ResponseMessageError): pass
class ErrorWrongServerVersion(ResponseMessageError): pass
class ErrorWrongServerVersionDelegate(ResponseMessageError): pass


# Microsoft recommends to cache the autodiscover data around 24 hours and perform autodiscover
# immediately following certain error responses from EWS. See more at
# http://blogs.msdn.com/b/mstehle/archive/2010/11/09/ews-best-practices-use-autodiscover.aspx

ERRORS_REQUIRING_REAUTODISCOVER = (
    ErrorAutoDiscoverFailed,
    ErrorConnectionFailed,
    ErrorIncorrectSchemaVersion,
    ErrorInvalidCrossForestCredentials,
    ErrorInvalidIdReturnedByResolveNames,
    ErrorInvalidNetworkServiceContext,
    ErrorMailboxMoveInProgress,
    ErrorMailboxMoveInProgress,
    ErrorMailboxStoreUnavailable,
    ErrorNameResolutionNoMailbox,
    ErrorNameResolutionNoResults,
    ErrorNotEnoughMemory,
)



from .credentials import DELEGATE, IMPERSONATION, Credentials
from .account import Account
from .autodiscover import discover
from .configuration import Configuration


def close_connections():
    from .autodiscover import close_connections as close_autodiscover_connections
    from .protocol import close_connections as close_protocol_connections
    close_autodiscover_connections()
    close_protocol_connections()



"""
Autodiscover is a Microsoft method for automatically getting the hostname of the Exchange server and the server
version of the server holding the email address using only the email address and password of the user (and possibly
User Principal Name). The protocol for autodiscovering an email address is described in detail in
http://msdn.microsoft.com/en-us/library/hh352638(v=exchg.140).aspx. Handling error messages is described here:
http://msdn.microsoft.com/en-us/library/office/dn467392(v=exchg.150).aspx. This is not fully implemented.

WARNING: We are taking many shortcuts here, like assuming SSL and following 302 Redirects automatically.
If you have problems autodiscovering, start by doing an official test at https://testconnectivity.microsoft.com
"""

# TODO: According to Microsoft, we may cache the URL of the autodiscover service forever, or until it stops responding.
# My previous experience with Exchange products in mind, I'm not sure if I should trust that advice. But it could save
# some valuable seconds every time we start a new connection to a known server. In any case, this info would require
# persistent storage.

from xml.etree.cElementTree import Element, tostring
import logging
from threading import Lock
import queue
from urllib import parse

import requests.exceptions
import dns.resolver

from .credentials import Credentials
from .version import API_VERSIONS
from .errors import AutoDiscoverFailed, AutoDiscoverRedirect, TransportError, RedirectError, ErrorNonExistentMailbox
from .protocol import Protocol
from . import transport
from .util import get_xml_attr, set_xml_attr, to_xml, is_xml, post_ratelimited, get_redirect_url

log = logging.getLogger(__name__)

REQUEST_NS = 'http://schemas.microsoft.com/exchange/autodiscover/outlook/requestschema/2006'
AUTODISCOVER_NS = 'http://schemas.microsoft.com/exchange/autodiscover/outlook/responseschema/2006'
ERROR_NS = 'http://schemas.microsoft.com/exchange/autodiscover/responseschema/2006'
RESPONSE_NS = 'http://schemas.microsoft.com/exchange/autodiscover/outlook/responseschema/2006a'

REQUEST_TIMEOUT = 10  # Seconds

# Used to cache the autoconfigure URL for a specific hostname
_autodiscover_cache = {}
_autodiscover_cache_lock = Lock()


def close_connections():
    for domain, cached_values in _autodiscover_cache.items():
        log.debug('Domain %s: Closing sessions', domain)
        autodiscover_protocol = cached_values[2]
        autodiscover_protocol.close()


def discover(email, credentials):
    """
    Performs the autodiscover dance and returns a Protocol on success. The autodiscover and EWs server might not be
    the same, so we use a different Protocol to do the autodiscover request, and return a hopefully-cached Protocol
    to the callee.
    """
    log.debug('Attempting autodiscover on email %s', email)
    assert isinstance(credentials, Credentials)
    domain = get_domain(email)
    # Use lock to guard against multiple threads competing to cache information
    if domain in _autodiscover_cache:
        # Python dict() is thread safe, so accessing _autodiscover_cache without a lock should be OK
        hostname, has_ssl, protocol = _autodiscover_cache[domain]
        assert isinstance(protocol, AutodiscoverProtocol)
        log.debug('Cache hit for domain %s: %s', domain, hostname)
        try:
            # This is the main path when the cache is primed
            primary_smtp_address, protocol = _autodiscover_quick(hostname=hostname, credentials=credentials,
                                                                 email=email, has_ssl=has_ssl,
                                                                 autodiscover_protocol=protocol)
            assert primary_smtp_address
            assert isinstance(protocol, Protocol)
            return primary_smtp_address, protocol
        except AutoDiscoverRedirect as e:
            log.debug('%s redirects to %s', email, e.redirect_email)
            if email.lower() == e.redirect_email.lower():
                raise AutoDiscoverFailed('Redirect to same email address: %s' % email) from e
            # Start over with the new email address
            return discover(e.redirect_email, credentials)
        # This is unreachable

    log.debug('Waiting for _autodiscover_cache_lock')
    with _autodiscover_cache_lock:
        log.debug('_autodiscover_cache_lock acquired')
        # Don't recurse while holding the lock!
        if domain in _autodiscover_cache:
            # Cache was primed by some other thread while we were waiting for the lock.
            log.debug('Cache filled for domain %s while we were waiting', domain)
        else:
            log.debug('Cache miss for domain %s', domain)
            log.debug('Cache contents: %s', _autodiscover_cache)
            try:
                # This eventually fills the cache in _autodiscover_hostname
                primary_smtp_address, protocol = _try_autodiscover(hostname=domain, credentials=credentials,
                                                                   email=email)
                assert primary_smtp_address
                assert isinstance(protocol, Protocol)
                return primary_smtp_address, protocol
            except AutoDiscoverRedirect as e:
                if email.lower() == e.redirect_email.lower():
                    raise AutoDiscoverFailed('Redirect to same email address: %s' % email) from e
                log.debug('%s redirects to %s', email, e.redirect_email)
                email = e.redirect_email
            finally:
                log.debug('Releasing_autodiscover_cache_lock')
    # We fell out of the with statement, so either cache was filled by someone else, or autodiscover redirected us to
    # another email address. Start over.
    return discover(email, credentials)


def _try_autodiscover(hostname, credentials, email):
    # Implements the full chain of autodiscover server discovery attempts. Tries to return autodiscover data from the
    # final host.
    try:
        return _autodiscover_hostname(hostname=hostname, credentials=credentials, email=email)
    except RedirectError as e:
        return _try_autodiscover(e.server, credentials, email)
    except AutoDiscoverFailed:
        try:
            return _autodiscover_hostname(hostname='autodiscover.%s' % hostname, credentials=credentials, email=email)
        except RedirectError as e:
            return _try_autodiscover(e.server, credentials, email)
        except AutoDiscoverFailed:
            try:
                return _autodiscover_hostname(hostname='autodiscover.%s' % hostname, credentials=credentials,
                                              email=email, has_ssl=False)
            except RedirectError as e:
                return _try_autodiscover(e.server, credentials, email)
            except AutoDiscoverFailed:
                try:
                    hostname_from_dns = _get_canonical_name(hostname='autodiscover.%s' % hostname)
                    if not hostname_from_dns:
                        hostname_from_dns = _get_hostname_from_srv(hostname='autodiscover.%s' % hostname)
                    # Start over with new hostname
                    return _try_autodiscover(hostname=hostname_from_dns, credentials=credentials, email=email)
                except AutoDiscoverFailed:
                    hostname_from_dns = _get_hostname_from_srv(hostname='_autodiscover._tcp.%s' % hostname)
                    # Start over with new hostname
                    return _try_autodiscover(hostname=hostname_from_dns, credentials=credentials, email=email)


def _autodiscover_hostname(hostname, credentials, email, has_ssl=True, auth_type=None):
    # Tries to get autodiscover data on a specific host. If we are HTTP redirected, we restart the autodiscover dance on
    # the new host.
    scheme = 'https' if has_ssl else 'http'
    url = '%s://%s/Autodiscover/Autodiscover.xml' % (scheme, hostname)
    log.debug('Trying autodiscover on %s', url)
    if not auth_type:
        try:
            auth_type = _get_autodiscover_auth_type(hostname=hostname, url=url, has_ssl=has_ssl, email=email)
        except RedirectError as e:
            log.debug(e)
            redirect_url, redirect_hostname, redirect_has_ssl = e.url, e.server, e.has_ssl
            log.debug('We were redirected to %s', redirect_url)
            canonical_hostname = _get_canonical_name(redirect_hostname)
            if canonical_hostname:
                log.debug('Canonical hostname is %s', canonical_hostname)
                redirect_hostname = canonical_hostname
            # Try the process on the new host, without 'www'. This is beyond the autodiscover protocol
            if redirect_hostname.startswith('www.'):
                redirect_hostname = redirect_hostname[4:]
            if redirect_hostname == hostname:
                log.debug('We were redirected to the same host')
                raise AutoDiscoverFailed('We were redirected to the same host') from e
            raise RedirectError(url=None, server=redirect_hostname, has_ssl=redirect_has_ssl) from e

    autodiscover_protocol = AutodiscoverProtocol(url=url, has_ssl=has_ssl, credentials=credentials, auth_type=auth_type)
    r = _get_autodiscover_response(protocol=autodiscover_protocol, url=url, email=email, verify=True)
    if r.status_code == 302:
        redirect_url, server, has_ssl = get_redirect_url(r, hostname, has_ssl)
        log.debug('We were redirected to %s', redirect_url)
        # Don't raise RedirectError here because we need to pass the ssl and auth_type data
        return _autodiscover_hostname(server, credentials, email, has_ssl=has_ssl, auth_type=None)
    domain = get_domain(email)
    try:
        server, has_ssl, ews_url, ews_auth_type, primary_smtp_address = _parse_response(r.text)
        if not primary_smtp_address:
            primary_smtp_address = email
    except (ErrorNonExistentMailbox, AutoDiscoverRedirect):
        # These are both valid responses from an autodiscover server, showing that we have found the correct
        # server for the original domain. Fill cache before re-raising
        log.debug('Adding cache entry for %s (hostname %s, has_ssl %s)' % (domain, hostname, has_ssl))
        _autodiscover_cache[domain] = hostname, has_ssl, autodiscover_protocol
        raise

    real_ews_auth_type = transport.get_service_authtype(server=server, has_ssl=has_ssl, ews_url=ews_url,
                                                        versions=API_VERSIONS)
    if ews_auth_type != real_ews_auth_type:
        log.debug('Autodiscover and real server disagree on auth method for %s (%s vs %s). Using server version',
                  email, ews_auth_type, real_ews_auth_type)
        ews_auth_type = real_ews_auth_type

    # Cache the final hostname of the autodiscover service so we don't need to autodiscover the same domain again
    log.debug('Adding cache entry for %s (hostname %s, has_ssl %s)' % (domain, hostname, has_ssl))
    _autodiscover_cache[domain] = hostname, has_ssl, autodiscover_protocol
    protocol = Protocol(has_ssl=has_ssl, ews_url=ews_url, credentials=credentials, ews_auth_type=ews_auth_type)
    return primary_smtp_address, protocol


def _autodiscover_quick(hostname, credentials, email, has_ssl, autodiscover_protocol):
    scheme = 'https' if has_ssl else 'http'
    url = '%s://%s/autodiscover/autodiscover.xml' % (scheme, hostname)
    r = _get_autodiscover_response(protocol=autodiscover_protocol, url=url, email=email, verify=True)
    server, has_ssl, ews_url, ews_auth_type, primary_smtp_address = _parse_response(r.text)
    if not primary_smtp_address:
        primary_smtp_address = email
    log.debug('Autodiscover success: %s may connect to %s as primary email %s', email, ews_url, primary_smtp_address)
    protocol = Protocol(has_ssl=has_ssl, ews_url=ews_url, credentials=credentials, ews_auth_type=ews_auth_type)
    return primary_smtp_address, protocol


def _get_autodiscover_auth_type(hostname, url, has_ssl, email, encoding='utf-8'):
    try:
        data = _get_autodiscover_payload(email=email, encoding=encoding)
        return transport.get_autodiscover_authtype(server=hostname, has_ssl=has_ssl, url=url, data=data,
                                                   timeout=REQUEST_TIMEOUT)
    except (TransportError, requests.exceptions.ConnectionError, requests.exceptions.Timeout,
            requests.exceptions.SSLError) as e:
        if isinstance(e, RedirectError):
            raise
        log.debug('Error guessing auth type: %s', e)
        raise AutoDiscoverFailed('Error guessing auth type: %s' % e) from e


def _get_autodiscover_payload(email, encoding='utf-8'):
    # Builds a full Autodiscover XML request
    payload = Element('Autodiscover')
    payload.set('xmlns', REQUEST_NS)
    request = Element('Request')
    set_xml_attr(request, 'EMailAddress', email)
    set_xml_attr(request, 'AcceptableResponseSchema', RESPONSE_NS)
    payload.append(request)
    return ('<?xml version="1.0" encoding="%s"?>' % encoding).encode(encoding) + tostring(payload, encoding=encoding)


def _get_autodiscover_response(protocol, url, email, encoding='utf-8', verify=True):
    data = _get_autodiscover_payload(email=email, encoding=encoding)
    headers = {'Content-Type': 'text/xml; charset=%s' % encoding}
    try:
        # Rate-limiting is an issue with autodiscover if the same setup is hosting EWS and autodiscover and we just
        # hammered the server with requests. We allow redirects since some autodiscover servers will issue different
        # redirects depending on the POST data content.
        session = protocol.get_session()
        r, session = post_ratelimited(protocol=protocol, session=session, url=url, headers=headers, data=data,
                                      timeout=protocol.timeout, verify=verify, allow_redirects=True)
        protocol.release_session(session)
        log.debug('Response headers: %s', r.headers)
    except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as e:
        log.debug('Connection error on %s: %s', url, e)
        # Don't raise AutoDiscoverFailed here. Connection errors could just as well be a valid but misbehaving server.
        raise
    except RedirectError:
        raise
    except TransportError:
        log.debug('No access to %s using %s', url, protocol.ews_auth_type)
        raise AutoDiscoverFailed('No access to %s using %s' % (url, protocol.ews_auth_type))
    if r.status_code == 302:
        # Give caller a chance to re-do the request
        return r
    if r.status_code != 200:
        log.debug('%s returned HTTP %s', url, r.status_code)
        # raise an uncatched error for now, until we understand this failure case
        raise TransportError('%s returned HTTP %s' % (url, r.status_code))
    if not is_xml(r.text):
        # This is normal - e.g. a greedy webserver serving custom HTTP 404's as 200 OK
        log.debug('URL %s: This is not XML: %s', url, r.text[:1000])
        raise AutoDiscoverFailed('URL %s: This is not XML: %s' % (url, r.text[:1000]))
    return r


def _parse_response(response, encoding='utf-8'):
    # We could return lots more interesting things here
    # log.debug('Autodiscover response: %s', response)
    autodiscover = to_xml(response, encoding=encoding)
    resp = autodiscover.find('{%s}Response' % RESPONSE_NS)
    if resp is None:
        resp = autodiscover.find('{%s}Response' % ERROR_NS)
        error = resp.find('{%s}Error' % ERROR_NS)
        errorcode = get_xml_attr(error, '{%s}ErrorCode' % ERROR_NS)
        message = get_xml_attr(error, '{%s}Message' % ERROR_NS)
        if message in ('The e-mail address cannot be found.', "The email address can't be found."):
            raise ErrorNonExistentMailbox('The SMTP address has no mailbox associated with it')
        raise AutoDiscoverFailed('Unknown error %s: %s' % (errorcode, message))
    account = resp.find('{%s}Account' % RESPONSE_NS)
    action = get_xml_attr(account, '{%s}Action' % RESPONSE_NS)
    redirect_email = get_xml_attr(account, '{%s}RedirectAddr' % RESPONSE_NS)
    if action == 'redirectAddr' and redirect_email:
        # This is redirection to e.g. Office365
        raise AutoDiscoverRedirect(redirect_email)
    user = resp.find('{%s}User' % RESPONSE_NS)
    # AutoDiscoverSMTPAddress might not be present in the XML, so primary_smtp_address might be None. In this
    # case, the original email address IS the primary address
    primary_smtp_address = get_xml_attr(user, '{%s}AutoDiscoverSMTPAddress' % RESPONSE_NS)
    account_type = get_xml_attr(account, '{%s}AccountType' % RESPONSE_NS)
    assert account_type == 'email'
    protocols = account.findall('{%s}Protocol' % RESPONSE_NS)
    # There are three possible protocol types: EXCH, EXPR and WEB. EXPR is for EWS. See
    # http://blogs.technet.com/b/exchange/archive/2008/09/26/3406344.aspx
    for protocol in protocols:
        if get_xml_attr(protocol, '{%s}Type' % RESPONSE_NS) != 'EXPR':
            continue
        server = get_xml_attr(protocol, '{%s}Server' % RESPONSE_NS)
        has_ssl = True if get_xml_attr(protocol, '{%s}SSL' % RESPONSE_NS) == 'On' else False
        ews_url = get_xml_attr(protocol, '{%s}EwsUrl' % RESPONSE_NS)
        auth_package = get_xml_attr(protocol, '{%s}AuthPackage' % RESPONSE_NS)
        try:
            ews_auth_type = {
                'ntlm': transport.NTLM,
                'basic': transport.BASIC,
                'digest': transport.DIGEST,
                None: transport.NOAUTH,
            }[auth_package.lower() if auth_package else None]
        except KeyError:
            log.warning("Unknown auth package '%s'")
            ews_auth_type = transport.UNKNOWN
        log.debug('Primary SMTP:%s, EWS endpoint:%s, auth type:%s', primary_smtp_address, ews_url, ews_auth_type)
        assert server
        assert has_ssl in (True, False)
        assert ews_url
        assert ews_auth_type
        return server, has_ssl, ews_url, ews_auth_type, primary_smtp_address
    raise AutoDiscoverFailed('Invalid AutoDiscover response: %s' % response)


def _get_canonical_name(hostname):
    log.debug('Attempting to get canonical name for %s', hostname)
    resolver = dns.resolver.Resolver()
    resolver.timeout = REQUEST_TIMEOUT
    try:
        canonical_name = resolver.query(hostname).canonical_name.to_unicode().rstrip('.')
    except dns.resolver.NXDOMAIN:
        log.debug('Nonexistent domain %s', hostname)
        return None
    if canonical_name != hostname:
        log.debug('%s has canonical name %s', hostname, canonical_name)
        return canonical_name
    return None


def _get_hostname_from_srv(hostname):
    # May return e.g.:
    #   canonical name = mail.ucl.dk.
    #   service = 8 100 443 webmail.ucn.dk.
    # or throw dns.resolver.NoAnswer
    log.debug('Attempting to get SRV record on %s', hostname)
    resolver = dns.resolver.Resolver()
    resolver.timeout = REQUEST_TIMEOUT
    try:
        answers = resolver.query(hostname, 'SRV')
        for rdata in answers:
            try:
                vals = rdata.to_text().strip().rstrip('.').split(' ')
                priority, weight, port, svr = int(vals[0]), int(vals[1]), int(vals[2]), vals[3]
            except (ValueError, KeyError) as e:
                raise AutoDiscoverFailed('Incompatible SRV record for %s (%s)' % (hostname, rdata.to_text())) from e
            else:
                return svr
    except dns.resolver.NoNameservers as e:
        raise AutoDiscoverFailed('No name servers for %s' % hostname) from e
    except dns.resolver.NoAnswer as e:
        raise AutoDiscoverFailed('No SRV record for %s' % hostname) from e
    except dns.resolver.NXDOMAIN as e:
        raise AutoDiscoverFailed('Nonexistent domain %s' % hostname) from e


def get_domain(email):
    try:
        return email.split('@')[1].lower().strip()
    except (IndexError, AttributeError) as e:
        raise ValueError("'%s' is not a valid email" % email) from e

POOLSIZE = 4


class AutodiscoverProtocol(Protocol):
    # Dummy class for post_ratelimited which implements the bare essentials
    SESSION_POOLSIZE = 1

    def __init__(self, url, has_ssl, credentials, auth_type):
        assert isinstance(credentials, Credentials)
        self.server = parse.urlparse(url).hostname.lower()
        self.credentials = credentials
        self.ews_url = url
        self.ews_auth_type = auth_type
        self.has_ssl = has_ssl
        self.timeout = REQUEST_TIMEOUT
        self._session_pool = queue.LifoQueue(maxsize=POOLSIZE)
        for i in range(POOLSIZE):
            self._session_pool.put(self.create_session(), block=False)



from parser import expr
from token import NAME, EQEQUAL, NOTEQUAL, GREATEREQUAL, LESSEQUAL, LESS, GREATER, STRING, LPAR, RPAR, NEWLINE, \
    ENDMARKER
from symbol import and_test, or_test, not_test, comparison, eval_input, sym_name
from xml.etree.cElementTree import Element, ElementTree, tostring
import logging
from threading import Lock

from .ewsdatetime import EWSTimeZone

log = logging.getLogger(__name__)

_source_cache = dict()
_source_cache_lock = Lock()


class Restriction:
    """
    Implements an EWS Restriction type.

    """
    def __init__(self, source):
        """
        source is a search expression in Python syntax. EWS Item fieldnames may be spelled with a colon (:). They will
        be escaped as underscores (_) since colons are not allowed in Python identifiers. Example:

              calendar:Start > '2009-01-15T13:45:56Z' and ( not calendar:Subject == 'EWS Test' )
        """
        with _source_cache_lock:
            # Something within the parser module seems to be deadlocking. Wrap in lock
            if source not in _source_cache:
                _source_cache[source] = self.parse_source(source).getroot()
        self.xml = _source_cache[source]

    def parse_source(self, source):
        """
        Takes a string and returns an XML tree.

        """
        log.debug('Parsing source: %s', source)

        # Make the syntax of the expression legal Python syntax. Play safe and only do this for known field URI prefixes
        for prefix in ('conversation:', 'postitem:', 'distributionlist:', 'contacts:', 'task:', 'calendar:',
                       'meetingRequest:', 'meeting:', 'message:', 'item:', 'folder:'):
            new = prefix.replace(':', '_')
            source = source.replace(prefix, new)

        st = expr(source).tolist()
        etree = ElementTree(self.parse_syntaxtree(st))
        # etree.register_namespace('t', 'http://schemas.microsoft.com/exchange/services/2006/messages')
        # etree.register_namespace('m', 'http://schemas.microsoft.com/exchange/services/2006/types')
        log.debug('Source parsed')
        return etree

    def parse_syntaxtree(self, slist):
        """
        Takes a Python syntax tree containing a search restriction expression and returns the tree as EWS-formatted XML
        """
        key = slist[0]
        if isinstance(slist[1], list):
            if len(slist) == 2:
                # Let nested 2-element lists pass transparently
                return self.parse_syntaxtree(slist[1])
            else:
                if key == or_test:
                    e = Element('t:Or')
                    for item in [self.parse_syntaxtree(l) for l in slist[1:]]:
                        if item is not None:
                            e.append(item)
                    return e
                if key == and_test:
                    e = Element('t:And')
                    for item in [self.parse_syntaxtree(l) for l in slist[1:]]:
                        if item is not None:
                            e.append(item)
                    return e
                if key == not_test:
                    e = Element('t:Not')
                    for item in [self.parse_syntaxtree(l) for l in slist[1:]]:
                        if item is not None:
                            e.append(item)
                    return e
                if key == comparison:
                    op = self.parse_syntaxtree(slist[2])
                    field = self.parse_syntaxtree(slist[1])
                    constant = self.parse_syntaxtree(slist[3])
                    op.append(field)
                    if op.tag == 't:Contains':
                        op.append(constant)
                    else:
                        uriorconst = Element('t:FieldURIOrConstant')
                        uriorconst.append(constant)
                        op.append(uriorconst)
                    return op
                if key == eval_input:
                    e = Element('m:Restriction')
                    for item in [self.parse_syntaxtree(l) for l in slist[1:]]:
                        if item is not None:
                            e.append(item)
                    return e
                raise ValueError('Unknown element type: %s %s' % (key, sym_name[key]))
        else:
            val = slist[1]
            if key == NAME:
                if val in ('and', 'or', 'not'):
                    return None
                if val == 'in':
                    return Element('t:Contains', ContainmentMode='Substring', ContainmentComparison='Exact')
                return Element('t:FieldURI', FieldURI=val.replace('_', ':'))  # Switch back to correct spelling
            if key == EQEQUAL:
                return Element('t:IsEqualTo')
            if key == NOTEQUAL:
                return Element('t:IsNotEqualTo')
            if key == GREATEREQUAL:
                return Element('t:IsGreaterThanOrEqualTo')
            if key == LESSEQUAL:
                return Element('t:IsLessThanOrEqualTo')
            if key == LESS:
                return Element('t:IsLessThan')
            if key == GREATER:
                return Element('t:IsGreaterThan')
            if key == STRING:
                return Element('t:Constant', Value=val.strip('"\''))  # This is a string, so strip single/double quotes
            if key in (LPAR, RPAR, NEWLINE, ENDMARKER):
                return None
            raise ValueError('Unknown token type: %s %s' % (key, val))

    @classmethod
    def from_params(cls, folder_id, start=None, end=None, categories=None):
        # Builds a search expression string and returns the equivalent XML tree ready for inclusion in an EWS request
        if not (start or end or categories):
            return None
        search_expr = []
        tz = EWSTimeZone.timezone('UTC')
        if start:
            search_expr.append('%s:End > "%s"' % (folder_id, start.astimezone(tz).ewsformat()))
        if end:
            search_expr.append('%s:Start < "%s"' % (folder_id, end.astimezone(tz).ewsformat()))
        if categories:
            if len(categories) == 1:
                search_expr.append('item:Categories in "%s"' % categories[0])
            else:
                expr2 = []
                for cat in categories:
                    # TODO Searching for items with multiple categories seems to be broken in EWS. 'And' operator
                    # returns no items, and searching for a list of categories doesn't work either.
                    expr2.append('item:Categories in "%s"' % cat)
                search_expr.append('( ' + ' or '.join(expr2) + ' )')
        expr_str = ' and '.join(search_expr)
        return cls(expr_str)

    def __str__(self):
        """
        Prints the XML syntax tree
        """
        return tostring(self.xml).decode('utf-8')



"""
Implement a selection of EWS services.

Exchange is very picky about things like the order of XML elements in SOAP requests, so we need to generate XML
automatically instead of taking advantage of Python SOAP libraries and the WSDL file.

Exchange EWS references:
    - 2007: http://msdn.microsoft.com/en-us/library/bb409286(v=exchg.80).aspx
    - 2010: http://msdn.microsoft.com/en-us/library/bb409286(v=exchg.140).aspx
    - 2013: http://msdn.microsoft.com/en-us/library/bb409286(v=exchg.150).aspx
"""

import logging
import itertools
from xml.etree.cElementTree import Element, tostring
from xml.parsers.expat import ExpatError
import traceback

from . import errors
from .errors import EWSWarning, TransportError, SOAPError, ErrorTimeoutExpired, ErrorBatchProcessingStopped, \
    ErrorQuotaExceeded, ErrorCannotDeleteObject, ErrorCreateItemAccessDenied, ErrorFolderNotFound, \
    ErrorNonExistentMailbox, ErrorMailboxStoreUnavailable, ErrorImpersonateUserDenied, ErrorInternalServerError, \
    ErrorInternalServerTransientError, ErrorNoRespondingCASInDestinationSite, ErrorImpersonationFailed, \
    ErrorMailboxMoveInProgress, ErrorAccessDenied, ErrorConnectionFailed, RateLimitError, ErrorServerBusy, \
    ErrorTooManyObjectsOpened, ErrorInvalidLicense, ErrorInvalidSchemaVersionForMailboxVersion
from .transport import wrap, SOAPNS, TNS, MNS, ENS
from .util import chunkify, set_xml_attr, get_xml_attr, to_xml, post_ratelimited

log = logging.getLogger(__name__)


# Shape enums
IdOnly = 'IdOnly'
# This doesn't actually get all properties in FindItem, just the "first-class" ones. See
#    http://msdn.microsoft.com/en-us/library/office/dn600367(v=exchg.150).aspx
AllProperties = 'AllProperties'

# Traversal enums
SHALLOW = 'Shallow'
DEEP = 'Deep'
SOFTDELETED = 'SoftDeleted'

ElementType = type(Element('x'))  # Type is auto-generated inside cElementTree


class EWSService:
    SERVICE_NAME = None
    element_container_name = None
    extra_element_names = []

    def __init__(self, protocol, element_name=None):
        self.protocol = protocol
        self.element_name = element_name

    def payload(self, version, account, *args, **kwargs):
        return wrap(content=self._get_payload(*args, **kwargs), version=version, account=account)

    def _get_payload(self, *args, **kwargs):
        raise NotImplementedError()

    def _get_elements(self, payload, account=None):
        assert isinstance(payload, ElementType)
        try:
            response = self._get_response_xml(payload=payload, account=account)
            return self._get_elements_in_response(response=response)
        except (ErrorQuotaExceeded, ErrorCannotDeleteObject, ErrorCreateItemAccessDenied, ErrorTimeoutExpired,
                ErrorFolderNotFound, ErrorNonExistentMailbox, ErrorMailboxStoreUnavailable, ErrorImpersonateUserDenied,
                ErrorInternalServerError, ErrorInternalServerTransientError, ErrorNoRespondingCASInDestinationSite,
                ErrorImpersonationFailed, ErrorMailboxMoveInProgress, ErrorAccessDenied, ErrorConnectionFailed,
                RateLimitError, ErrorServerBusy, ErrorTooManyObjectsOpened, ErrorInvalidLicense):
            # These are known and understood, and don't require a backtrace
            # TODO: ErrorTooManyObjectsOpened means there are too many connections to the database. We should be able to
            # act on this by lowering the self.protocol connection pool size.
            raise
        except Exception:
            # This may run from a thread pool, which obfuscates the stack trace. Print trace immediately.
            log.warning('EWS %s, account %s: Exception in _get_elements: %s', self.protocol.ews_url, account,
                        traceback.format_exc(20))
            raise

    def _get_response_xml(self, payload, account=None):
        # Takes an XML tree and returns SOAP payload as an XML tree
        assert isinstance(payload, ElementType)
        # Microsoft really doesn't want to make our lives easy. The server may report one version in our initial version
        # guessing tango, but then the server may decide that any arbitrary legacy backend server may actually process
        # the request for an account. Prepare to handle ErrorInvalidSchemaVersionForMailboxVersion errors and set the
        # server version per-account.
        from .version import API_VERSIONS, Version
        hint = account.version.api_version if account else self.protocol.version.api_version
        api_versions = [hint] + [v for v in API_VERSIONS if v != hint]
        for api_version in api_versions:
            session = self.protocol.get_session()
            soap_payload = wrap(content=payload, version=api_version, account=account)
            r, session = post_ratelimited(
                protocol=self.protocol,
                session=session,
                url=self.protocol.ews_url,
                headers=None,
                data=soap_payload,
                timeout=self.protocol.timeout,
                verify=True,
                allow_redirects=False)
            self.protocol.release_session(session)
            try:
                soap_response_payload = to_xml(r.text, encoding=r.encoding or 'utf-8')
            except ExpatError as e:
                raise SOAPError('SOAP response is not XML: %s' % e) from e
            try:
                res = self._get_soap_payload(soap_response=soap_response_payload)
            except ErrorInvalidSchemaVersionForMailboxVersion:
                assert account # This should never happen for non-account services
                # The guessed server version is wrong for this account. Try the next version
                log.warning('API version %s was invalid for account %s', api_version, account)
                continue
            if account and account.version.api_version != api_version:
                # The api_version that worked was different than our hint. Set new version for account
                log.info('New API version for account %s (%s -> %s)', account, account.version.api_version, api_version)
                account.version = Version.from_response(requested_api_version=api_version, response=r)
            return res
        raise ErrorInvalidSchemaVersionForMailboxVersion('Tried versions %s but all were invalid for account %s' %
                                                         (api_versions, account))

    def _get_soap_payload(self, soap_response):
        log_prefix = 'EWS %s, service %s' % (self.protocol.ews_url, self.SERVICE_NAME)
        assert isinstance(soap_response, ElementType)
        body = soap_response.find('{%s}Body' % SOAPNS)
        if body is None:
            raise TransportError('No Body element in SOAP response')
        response = body.find('{%s}%sResponse' % (MNS, self.SERVICE_NAME))
        if response is None:
            fault = body.find('{%s}Fault' % SOAPNS)
            if fault is None:
                raise SOAPError('Unknown SOAP response: %s' % tostring(body))
            self._raise_soap_errors(fault=fault)  # Will throw SOAPError
        response_messages = response.find('{%s}ResponseMessages' % MNS)
        if response_messages is None:
            raise TransportError('%s: No ResponseMessages element in response' % log_prefix)
        return response_messages.findall('{%s}%sResponseMessage' % (MNS, self.SERVICE_NAME))

    def _raise_soap_errors(self, fault):
        assert isinstance(fault, ElementType)
        log_prefix = 'EWS %s, service %s' % (self.protocol.ews_url, self.SERVICE_NAME)
        # Fault: See http://www.w3.org/TR/2000/NOTE-SOAP-20000508/#_Toc478383507
        faultcode = get_xml_attr(fault, 'faultcode')
        faultstring = get_xml_attr(fault, 'faultstring')
        faultactor = get_xml_attr(fault, 'faultactor')
        detail = fault.find('detail')
        if detail is not None:
            code, msg = None, None
            if detail.find('{%s}ResponseCode' % ENS) is not None:
                code = get_xml_attr(detail, '{%s}ResponseCode' % ENS)
            if detail.find('{%s}Message' % ENS) is not None:
                msg = get_xml_attr(detail, '{%s}Message' % ENS)
            try:
                raise vars(errors)[code](msg)
            except KeyError:
                detail = '%s: code: %s msg: %s (%s)' % (log_prefix, code, msg, tostring(detail))
        try:
            raise vars(errors)[faultcode](faultstring)
        except KeyError:
            pass
        raise SOAPError('SOAP error code: %s string: %s actor: %s detail: %s' % (
            faultcode, faultstring, faultactor, detail))

    def _get_element_container(self, message, name=None):
        assert isinstance(message, ElementType)
        # ResponseClass: See http://msdn.microsoft.com/en-us/library/aa566424(v=EXCHG.140).aspx
        response_class = message.get('ResponseClass')
        # ResponseCode, MessageText: See http://msdn.microsoft.com/en-us/library/aa580757(v=EXCHG.140).aspx
        response_code = get_xml_attr(message, '{%s}ResponseCode' % MNS)
        msg_text = get_xml_attr(message, '{%s}MessageText' % MNS)
        msg_xml = get_xml_attr(message, '{%s}MessageXml' % MNS)
        if response_class == 'Success' and response_code == 'NoError':
            if not name:
                return True
            container = message.find(name)
            if container is None:
                raise TransportError('No %s elements in ResponseMessage (%s)' % (name, tostring(message)))
            return container
        if response_class == 'Warning':
            return self._raise_warnings(code=response_code, text=msg_text, xml=msg_xml)
        # rspclass == 'Error', or 'Success' and not 'NoError'
        return self._raise_errors(code=response_code, text=msg_text, xml=msg_xml)

    def _raise_warnings(self, code, text, xml):
        try:
            return self._raise_errors(code=code, text=text, xml=xml)
        except ErrorBatchProcessingStopped as e:
            raise EWSWarning(e.value) from e

    @staticmethod
    def _raise_errors(code, text, xml):
        if code == 'NoError':
            return True
        if not code:
            raise TransportError('Empty ResponseCode in ResponseMessage (MessageText: %s, MessageXml: %s)' % (
                text, xml))
        try:
            # Raise the error corresponding to the ResponseCode
            raise vars(errors)[code](text)
        except KeyError as e:
            # Should not happen
            raise TransportError('Unknown ResponseCode in ResponseMessage: %s (MessageText: %s, MessageXml: %s)' % (
                code, text, xml)) from e

    def _get_elements_in_response(self, response):
        assert isinstance(response, list)
        elements = []
        for msg in response:
            assert isinstance(msg, ElementType)
            try:
                container = self._get_element_container(message=msg, name=self.element_container_name)
                if isinstance(container, ElementType):
                    elements.extend(self._get_elements_in_container(container=container))
                else:
                    elements.append((container, None))
            except (ErrorTimeoutExpired, ErrorBatchProcessingStopped):
                raise
            except EWSWarning as e:
                elements.append((False, '%s' % e.value))
                continue
        return elements

    def _get_elements_in_container(self, container):
        elems = container.findall(self.element_name)
        for element_name in self.extra_element_names:
            elems.extend(container.findall(element_name))
        return elems


class EWSAccountService(EWSService):
    def call(self, account, **kwargs):
        raise NotImplementedError()


class EWSFolderService(EWSService):
    def call(self, folder, **kwargs):
        raise NotImplementedError()


class PagingEWSService(EWSService):
    def _paged_call(self, **kwargs):
        # TODO This is awkward. The function must work with _get_payload() of both folder- and account-based services
        account = kwargs['folder'].account if 'folder' in kwargs else kwargs['account']
        log_prefix = 'EWS %s, account %s, service %s' % (self.protocol.ews_url, account, self.SERVICE_NAME)
        elements = []
        offset = 0
        while True:
            log.debug('%s: Getting %s at offset %s', log_prefix, self.element_name, offset)
            kwargs['offset'] = offset
            payload = self._get_payload(**kwargs)
            response = self._get_response_xml(payload=payload, account=account)
            page, offset = self._get_page(response)
            if isinstance(page, ElementType):
                container = page.find(self.element_container_name)
                if container is None:
                    raise TransportError('No %s elements in ResponseMessage (%s)' % (self.element_container_name,
                                                                                     tostring(page)))
                elements.extend(self._get_elements_in_container(container=container))
            if not offset:
                break
        return elements

    def _get_page(self, response):
        assert len(response) == 1
        log_prefix = 'EWS %s, service %s' % (self.protocol.ews_url, self.SERVICE_NAME)
        rootfolder = self._get_element_container(message=response[0], name='{%s}RootFolder' % MNS)
        is_last_page = rootfolder.get('IncludesLastItemInRange').lower() in ('true', '0')
        offset = rootfolder.get('IndexedPagingOffset')
        if offset is None and not is_last_page:
            log.warning("Not last page in range, but Exchange didn't send a page offset. Assuming first page")
            offset = '1'
        next_offset = 0 if is_last_page else int(offset)
        if not int(rootfolder.get('TotalItemsInView')):
            assert next_offset == 0
            rootfolder = None
        log.debug('%s: Got page with next offset %s (last_page %s)', log_prefix, next_offset, is_last_page)
        return rootfolder, next_offset


class GetServerTimeZones(EWSService):
    SERVICE_NAME = 'GetServerTimeZones'
    element_container_name = '{%s}TimeZoneDefinitions' % MNS

    def __init__(self, *args, **kwargs):
        kwargs['element_name'] = '{%s}TimeZoneDefinition' % TNS
        super().__init__(*args, **kwargs)

    def call(self, **kwargs):
        if self.protocol.version.major_version < 14:
            raise NotImplementedError('%s is only supported for Exchange 2010 servers and later' % self.SERVICE_NAME)
        return self._get_elements(payload=self._get_payload(**kwargs))

    def _get_payload(self, returnfulltimezonedata=False):
        return Element('m:%s' % self.SERVICE_NAME, ReturnFullTimeZoneData=(
            'true' if returnfulltimezonedata else 'false'))

    def _get_elements_in_container(self, container):
        timezones = []
        timezonedefs = container.findall(self.element_name)
        for timezonedef in timezonedefs:
            tz_id = timezonedef.get('Id')
            name = timezonedef.get('Name')
            timezones.append((tz_id, name))
        return timezones


class EWSPooledService(EWSService):
    CHUNKSIZE = None

    def _pool_requests(self, account, payload_func, items):
        log.debug('Processing %s items in chunks of %s', len(items), self.CHUNKSIZE)
        # Chop items list into suitable pieces and let worker threads chew on the work. The order of the output result
        # list must be the same as the input id list, so the caller knows which status message belongs to which ID.
        func = lambda n: self._get_elements(account=account, payload=payload_func(n))
        return list(itertools.chain(*account.protocol.thread_pool.map(func, chunkify(items, self.CHUNKSIZE))))


class GetItem(EWSPooledService):
    """
    Take a list of (id, changekey) tuples and returns a list of items in stable order
    """
    CHUNKSIZE = 100
    SERVICE_NAME = 'GetItem'
    element_container_name = '{%s}Items' % MNS

    def call(self, folder, **kwargs):
        self.element_name = '{%s}%s' % (TNS, folder.item_model.ITEM_TYPE)
        return self._pool_requests(account=folder.account, payload_func=folder.get_xml, items=kwargs['ids'])


class CreateItem(EWSPooledService):
    """
    Takes folder and a list of items. Returns result of creation as a list of tuples (success[True|False],
    errormessage), in the same order as the input list.

    """
    CHUNKSIZE = 25
    SERVICE_NAME = 'CreateItem'
    element_container_name = '{%s}Items' % MNS

    def call(self, folder, **kwargs):
        self.element_name = '{%s}%s' % (TNS, folder.item_model.ITEM_TYPE)
        return self._pool_requests(account=folder.account, payload_func=folder.create_xml, items=kwargs['items'])


class DeleteItem(EWSPooledService):
    """
    Takes a folder and a list of (id, changekey) tuples. Returns result of deletion as a list of tuples
    (success[True|False], errormessage), in the same order as the input list.

    """
    CHUNKSIZE = 25
    SERVICE_NAME = 'DeleteItem'
    element_container_name = None  # DeleteItem doesn't return a response object, just status in XML attrs

    def call(self, folder, **kwargs):
        return self._pool_requests(account=folder.account, payload_func=folder.delete_xml, items=kwargs['ids'])


class UpdateItem(EWSPooledService):
    CHUNKSIZE = 25
    SERVICE_NAME = 'UpdateItem'
    element_container_name = '{%s}Items' % MNS

    def call(self, folder, **kwargs):
        self.element_name = '{%s}%s' % (TNS, folder.item_model.ITEM_TYPE)
        return self._pool_requests(account=folder.account, payload_func=folder.update_xml, items=kwargs['items'])


class FindItem(PagingEWSService, EWSFolderService):
    """
    Gets all items for 'account' in folder 'folder_id', optionally expanded with 'additional_fields' Element,
    optionally restricted by a Restriction definition.
    """
    SERVICE_NAME = 'FindItem'
    element_container_name = '{%s}Items' % TNS

    def call(self, folder, **kwargs):
        self.element_name = '{%s}%s' % (TNS, folder.item_model.ITEM_TYPE)
        return self._paged_call(folder=folder, **kwargs)

    def _get_payload(self, folder, additional_fields=None, restriction=None, shape=IdOnly, depth=SHALLOW, offset=0):
        log.debug(
            'Finding %s items for %s extra fields %s restriction %s shape %s offset %s',
            folder.DISTINGUISHED_FOLDER_ID,
            folder.account,
            additional_fields,
            restriction,
            shape,
            offset,
        )
        finditem = Element('m:%s' % self.SERVICE_NAME, Traversal=SHALLOW)
        itemshape = Element('m:ItemShape')
        set_xml_attr(itemshape, 't:BaseShape', shape)
        if additional_fields:
            additionalproperties = Element('t:AdditionalProperties')
            for field_uri in additional_fields:
                additionalproperties.append(Element('t:FieldURI', FieldURI=field_uri))
            itemshape.append(additionalproperties)
        finditem.append(itemshape)
        indexedpageviewitem = Element('m:IndexedPageItemView', Offset=str(offset), BasePoint='Beginning')
        finditem.append(indexedpageviewitem)
        if restriction:
            finditem.append(restriction.xml)
        parentfolderids = Element('m:ParentFolderIds')
        parentfolderids.append(folder.folderid_xml())
        finditem.append(parentfolderids)
        return finditem


class FindFolder(PagingEWSService, EWSFolderService):
    """
    Gets a list of folders belonging to an account.
    """
    SERVICE_NAME = 'FindFolder'
    element_container_name = '{%s}Folders' % TNS
    # See http://msdn.microsoft.com/en-us/library/aa564009(v=exchg.150).aspx
    extra_element_names = [
        '{%s}CalendarFolder' % TNS,
        '{%s}ContactsFolder' % TNS,
        '{%s}SearchFolder' % TNS,
        '{%s}TasksFolder' % TNS,
    ]

    def __init__(self, *args, **kwargs):
        kwargs['element_name'] = '{%s}Folder' % TNS
        super().__init__(*args, **kwargs)

    def call(self, folder, **kwargs):
        return self._paged_call(folder=folder, **kwargs)

    def _get_payload(self, folder, additional_fields=None, shape=IdOnly, depth=DEEP, offset=0):
        log.debug(
            'Getting folders for %s, root:%s, extra fields:%s, shape:%s',
            folder.account,
            folder.name,
            additional_fields,
            shape
        )
        findfolder = Element('m:%s' % self.SERVICE_NAME, Traversal=depth)
        foldershape = Element('m:FolderShape')
        set_xml_attr(foldershape, 't:BaseShape', shape)
        if additional_fields:
            additionalproperties = Element('t:AdditionalProperties')
            for field_uri in additional_fields:
                additionalproperties.append(Element('t:FieldURI', FieldURI=field_uri))
            foldershape.append(additionalproperties)
        findfolder.append(foldershape)
        if folder.account.protocol.version.major_version >= 14:
            indexedpageviewitem = Element('m:IndexedPageFolderView', Offset=str(offset), BasePoint='Beginning')
            findfolder.append(indexedpageviewitem)
        else:
            assert offset == 0, 'Offset is %s' % offset
        parentfolderids = Element('m:ParentFolderIds')
        parentfolderids.append(folder.folderid_xml())
        findfolder.append(parentfolderids)
        return findfolder


class GetFolder(EWSFolderService):
    SERVICE_NAME = 'GetFolder'
    element_container_name = '{%s}Folders' % MNS
    # See http://msdn.microsoft.com/en-us/library/aa564009(v=exchg.150).aspx
    extra_element_names = [
        '{%s}CalendarFolder' % TNS,
        '{%s}ContactsFolder' % TNS,
        '{%s}SearchFolder' % TNS,
        '{%s}TasksFolder' % TNS,
    ]

    def __init__(self, *args, **kwargs):
        kwargs['element_name'] = '{%s}Folder' % TNS
        super().__init__(*args, **kwargs)

    def call(self, folder, **kwargs):
        return self._get_elements(payload=self._get_payload(folder, **kwargs), account=folder.account)

    def _get_payload(self, folder, additional_fields=None, shape=IdOnly):
        log.debug(
            'Getting folders for %s, folder:%s, extra fields:%s, shape:%s',
            folder.account,
            folder.name,
            additional_fields,
            shape
        )
        getfolder = Element('m:%s' % self.SERVICE_NAME)
        foldershape = Element('m:FolderShape')
        set_xml_attr(foldershape, 't:BaseShape', shape)
        if additional_fields:
            additionalproperties = Element('t:AdditionalProperties')
            for field_uri in additional_fields:
                additionalproperties.append(Element('t:FieldURI', FieldURI=field_uri))
            foldershape.append(additionalproperties)
        getfolder.append(foldershape)
        folderids = Element('m:FolderIds')
        folderids.append(folder.folderid_xml())
        getfolder.append(folderids)
        return getfolder


class ResolveNames(EWSAccountService):
    SERVICE_NAME = 'ResolveNames'
    element_container_name = '{%s}ResolutionSet' % MNS

    def __init__(self, *args, **kwargs):
        kwargs['element_name'] = '{%s}Resolution' % TNS
        super().__init__(*args, **kwargs)

    def call(self, **kwargs):
        return self._get_elements(payload=self._get_payload(**kwargs))

    def _get_payload(self, unresolvedentries, returnfullcontactdata=False):
        payload = Element('m:%s' % self.SERVICE_NAME, ReturnFullContactData=(
            'true' if returnfullcontactdata else 'false'))
        assert len(unresolvedentries)
        for entry in unresolvedentries:
            set_xml_attr(payload, 'm:UnresolvedEntry', entry)
        return payload
